"""Base communication client support for hosting services.

Version Added:
    6.0:
    This replaces the hosting service code in the old
    :py:mod:`reviewboard.hostingsvcs.service` module.
"""

from __future__ import annotations

import logging
from typing import Any, Dict, Mapping, TYPE_CHECKING
from urllib.parse import urlparse

from django.utils.translation import gettext_lazy as _
from housekeeping import deprecate_non_keyword_only_args
from typing_extensions import TypedDict

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.hostingsvcs.base.client import HostingServiceClient

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import ClassVar

    from django.http import HttpRequest
    from django.urls import _AnyURL
    from django.utils.safestring import SafeString
    from paramiko import PKey
    from typelets.django.strings import StrOrPromise
    from typing_extensions import NotRequired, TypeAlias

    from reviewboard.hostingsvcs.base.forms import (
        BaseHostingServiceAuthForm,
        BaseHostingServiceRepositoryForm)
    from reviewboard.hostingsvcs.base.paginator import BasePaginator
    from reviewboard.hostingsvcs.models import HostingServiceAccount
    from reviewboard.hostingsvcs.repository import RemoteRepository
    from reviewboard.scmtools.core import Branch, Commit, SCMTool
    from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


class HostingServicePlan(TypedDict):
    """Information on a hosting service plan.

    Version Added:
        6.0
    """

    #: The display name of the plan.
    #:
    #: This is required for all plans.
    #:
    #: Type:
    #:     str
    name: StrOrPromise

    #: A templated value for the bug tracker field.
    #:
    #: The value will be filled in based on the cleaned bug tracker form
    #: data. It must take templated keys in the form of ``%(key_name)s``.
    #:
    #: Type:
    #:     str
    bug_tracker_field: NotRequired[str]

    #: Optional form used to configure repository settings.
    #:
    #: If not provided, a default form will be used.
    #:
    #: This value can also be set within :py:attr:`repository_plans`.
    #:
    #: Type:
    #:     type
    form: NotRequired[type[BaseHostingServiceRepositoryForm]]

    #: Templated values to set for model repository fields.
    #:
    #: Each key corresponds to a SCMTool ID or name, and each value to a
    #: mapping of model attribute names to templates for values to populate
    #: based on cleaned repository form data. It must take templated keys
    #: in the form of ``%(key_name)s``.
    #:
    #: Type:
    #:     dict
    repository_fields: NotRequired[SCMToRepositoryFields]


#: Credentials used to authenticate with a hosting service.
#:
#: Version Added:
#:     6.0
HostingServiceCredentials: TypeAlias = Dict[str, Any]


#: Information on fields in a repository.
#:
#: These will map model or form attributes to values or templates for
#: processing when saving a repository.
#:
#: Version Added:
#:     6.0
RepositoryFields: TypeAlias = Dict[str, Any]


#: A mapping of SCMTool IDs to repository fields.
#:
#: Version Added:
#:     6.0
SCMToRepositoryFields: TypeAlias = Mapping[str, RepositoryFields]


class BaseHostingService:
    """An interface to a hosting service for repositories and bug trackers.

    Subclasses are used to more easily configure repositories and to make use
    of third party APIs to perform special operations not otherwise usable by
    generic repositories.

    A hosting service can specify forms for authentication, repository, and bug
    tracker configuration.

    It can also provide a list of repository "plans" (such as public
    repositories, private repositories, or other types available to the hosting
    service), along with configuration specific to the plan. These plans will
    be available when configuring the repository.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.service` to
          :py:mod:`reviewboard.hostingsvcs.base.hosting_service` and
          renamed from ``HostingService`` to ``BaseHostingService``.
    """

    #: The unique ID of the hosting service.
    #:
    #: This should be lowercase, and only consist of the characters a-z, 0-9,
    #: ``_``, and ``-``.
    #:
    #: Version Added:
    #:     3.0.16:
    #:     This should now be set on all custom hosting services. It will be
    #:     required in Review Board 4.0.
    #:
    #: Type:
    #:     str
    hosting_service_id: ClassVar[str | None] = None

    #: The display name for the hosting service.
    #:
    #: Type:
    #:     str
    name: ClassVar[StrOrPromise | None] = None

    #: A list of available hosting service plans.
    #:
    #: Each plan may map to different configuration forms and saved state,
    #: helping subclasses support different APIs or data models provided
    #: by the server, and better helping users supply the information needed
    #: to configure the repository.
    #:
    #: Type:
    #:     list
    plans: ClassVar[Sequence[tuple[str, HostingServicePlan]] | None] = None

    #: Whether this service supports bug trackers.
    #:
    #: This should be set to ``True`` if the service provides its own bug
    #: tracker and the subclass implements all of the following methods:
    #:
    #: * :py:meth:`get_bug_tracker_field`
    #: * :py:meth:`get_bug_tracker_requires_username`
    #:
    #: Type:
    #:     bool
    supports_bug_trackers: ClassVar[bool] = False

    #: Whether this service supports post-commit requests.
    #:
    #: This should be set to ``True`` if the subclass implements all of the
    #: following methods:
    #:
    #: * :py:meth:`get_branches`
    #: * :py:meth:`get_change`
    #: * :py:meth:`get_commits`
    #:
    #: Type:
    #:     bool
    supports_post_commit: ClassVar[bool] = False

    #: Whether this service supports source code repositories.
    #:
    #: This should be set to ``True`` if the service provides a compatible
    #: source code repository and the subclass implements all of the
    #: following methods:
    #:
    #: * :py:meth:`get_file`
    #: * :py:meth:`get_file_exists`
    #: * :py:meth:`get_repository_fields`
    #:
    #: Several other optional methods may also be defined and used if this
    #: is set.
    #:
    #: Type:
    #:     bool
    supports_repositories: ClassVar[bool] = False

    #: Whether this service supports automatic association of SSH keys.
    #:
    #: This should be set to ``True`` if the service provides any way of
    #: automatically assigning an SSH key for a user for repository access,
    #: and the subclass implements the following methods:
    #:
    #: * :py:meth:`associate_ssh_key`.
    #: * :py:meth:`is_ssh_key_associated`.
    #:
    #: Type:
    #:     bool
    supports_ssh_key_association: ClassVar[bool] = False

    #: Whether this service supports two-factor authentication.
    #:
    #: This should be set to ``True`` if the service supports two-factor
    #: authentication when communicating with the API and the service
    #: implements support in :py:meth:`authorize`.
    #:
    #: Type:
    #:     bool
    supports_two_factor_auth: ClassVar[bool] = False

    #: Whether this service supports listing remote repositories.
    #:
    #: This should be set to ``True`` if the service supports looking up
    #: repositories via the API and the subclass implements all of the
    #: following methods:
    #:
    #: * :py:meth:`get_remote_repositories`
    #: * :py:meth:`get_remote_repository`
    #:
    #: Type:
    #:     bool
    supports_list_remote_repositories: ClassVar[bool] = False

    #: Whether this service provides repository hook instructions.
    #:
    #: This should be set to ``True`` if the subclass implements instructions'
    #: on configuring useful repository hooks in the following method:
    #:
    #: * :py:meth:`get_repository_hook_instructions`
    #:
    #: Type:
    #:     bool
    has_repository_hook_instructions: ClassVar[bool] = False

    #: Whether this service should be shown as an available option.
    #:
    #: This should be set to ``False`` when a service is no longer available
    #: to use, and should be hidden from repository configuration. The
    #: implementation can then be largely stubbed out. Users will see a
    #: message in the repository configuration page.
    #:
    #: Version Added:
    #:     3.0.17
    #:
    #: Type:
    #:     bool
    visible: ClassVar[bool] = True

    #: Whether this service can be self-hosted.
    #:
    #: This should be set to ``True`` if the service can be hosted at a
    #: custom URL.
    #:
    #: Type:
    #:     bool
    self_hosted: ClassVar[bool] = False

    #: Custom URL patterns to include for the hosting service.
    #:
    #: This can be used to create URLs for hooks or other utility purposes.
    #: These will be installed under:
    #:
    #: * ``/repos/<repository_id>/<hosting_service_id>/``
    #: * ``/s/<local_site>/repos/<repository_id>/<hosting_service_id>/``
    #:
    #: Type:
    #:     list
    repository_url_patterns: ClassVar[list[_AnyURL] | None] = None

    #: The utility class used to perform client functionality.
    #:
    #: Client classes are responsible for communicating with the hosting
    #: service at an API level. Instance of this hosting service will receive
    #: a client instance.
    #:
    #: Subclasses should set this to a subclass of
    #: :py:class:`HostingServiceClient`.
    #:
    #: Type:
    #:     type
    client_class: ClassVar[type[HostingServiceClient]] = HostingServiceClient

    #: Optional form used to configure authentication settings for an account.
    #:
    #: If not provided, a default form will be used.
    #:
    #: Type:
    #:     type
    auth_form: ClassVar[type[BaseHostingServiceAuthForm] | None] = None

    #: Whether usage of this hosting service requires authorization.
    #:
    #: This value can also be set within :py:attr:`repository_plans`.
    #:
    #: Type:
    #:     bool
    needs_authorization: ClassVar[bool] = False

    #: Optional form used to configure repository settings.
    #:
    #: If not provided, a default form will be used.
    #:
    #: This value can also be set within :py:attr:`repository_plans`.
    #:
    #: Type:
    #:     type
    form: ClassVar[type[BaseHostingServiceRepositoryForm] | None] = None

    #: Templated values to set for model repository fields.
    #:
    #: Each key corresponds to a SCMTool ID or name, and each value to a
    #: mapping of model attribute names to templates for values to populate
    #: based on cleaned repository form data. It must take templated keys
    #: in the form of ``%(key_name)s``.
    #:
    #: Type:
    #:     dict
    repository_fields: ClassVar[SCMToRepositoryFields] = {}

    #: A templated value for the bug tracker field.
    #:
    #: The value will be filled in based on the cleaned bug tracker form
    #: data. It must take templated keys in the form of ``%(key_name)s``.
    #:
    #: Type:
    #:     str
    bug_tracker_field: ClassVar[str | None] = None

    #: A list of SCMTools IDs or names that are supported by this service.
    #:
    #: This should contain a list of SCMTool IDs that this service can work
    #: with. For backwards-compatibility, it may instead contain a list of
    #: SCMTool names (corresponding to database registration names).
    #:
    #: This may also be specified per-plan in the :py:attr:`plans`.
    #:
    #: Version Changed:
    #:     3.0.16:
    #:     Added support for SCMTool IDs. A future version will deprecate
    #:     using SCMTool names here.
    #:
    #: Type:
    #:     list of str
    supported_scmtools: ClassVar[Sequence[str]] = []

    #: A list of SCMTool IDs that are visible when configuring the service.
    #:
    #: This should contain a list of SCMTool IDs that this service will show
    #: when configuring a repository. It can be used to offer continued
    #: legacy support for an SCMTool without offering it when creating new
    #: repositories. If not specified, all SCMTools listed
    #: in :py:attr:`supported_scmtools` are assumed to be visible.
    #:
    #: If explicitly set, this should always be equal to or a subset of
    #: :py:attr:`supported_scmtools`.
    #:
    #: This may also be specified per-plan in the :py:attr:`plans`.
    #:
    #: Version Added:
    #:     3.0.17
    #:
    #: Type:
    #:     list of str
    visible_scmtools: ClassVar[Sequence[str] | None] = None

    ######################
    # Instance variables #
    ######################

    #: The client instance for talking to the hosting service.
    #:
    #: Type:
    #:     HostingServiceClient
    client: HostingServiceClient

    def __init__(
        self,
        account: HostingServiceAccount,
    ) -> None:
        """Initialize the hosting service.

        Args:
            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The account to use with the service.
        """
        assert account
        self.account = account

        self.client = self.client_class(self)

    def is_authorized(self) -> bool:
        """Return whether or not the account is currently authorized.

        An account may no longer be authorized if the hosting service
        switches to a new API that doesn't match the current authorization
        records. This function will determine whether the account is still
        considered authorized.

        Returns:
            bool:
            Whether or not the associated account is authorized.
        """
        return False

    def get_password(self) -> str | None:
        """Return the raw password for this hosting service.

        Not all hosting services provide this, and not all would need it.
        It's primarily used when building a Subversion client, or other
        SCMTools that still need direct access to the repository itself.

        Returns:
            str:
            The password.
        """
        return None

    def is_ssh_key_associated(
        self,
        repository: Repository,
        key: PKey,
    ) -> bool:
        """Return whether or not the key is associated with the repository.

        If the given key is present amongst the hosting service's deploy keys
        for the given repository, then it is considered to be associated.

        Subclasses should implement this when the hosting service supports
        SSH key association.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the key must be associated with.

            key (paramiko.PKey):
                The key to check for association.

        Returns:
            bool:
            Whether or not the key is associated with the repository.

        Raises:
            reviewboard.hostingsvcs.errors.SSHKeyAssociationError:
                If an error occurred during communication with the hosting
                service.
        """
        raise NotImplementedError

    def associate_ssh_key(
        self,
        repository: Repository,
        key: PKey,
    ) -> None:
        """Associate an SSH key with a given repository.

        Subclasses should implement this when the hosting service supports
        SSH key association.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to associate the key with.

            key (paramiko.PKey):
                The key to add to the repository's list of deploy keys.

        Raises:
            reviewboard.hostingsvcs.errors.SSHKeyAssociationError:
                If an error occurred during key association.
        """
        raise NotImplementedError

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def authorize(
        self,
        *,
        username: str | None,
        password: str | None,
        hosting_url: (str | None) = None,
        credentials: (HostingServiceCredentials | None) = None,
        two_factor_auth_code: (str | None) = None,
        local_site_name: (str | None) = None,
        **kwargs,
    ) -> None:
        """Authorize an account for the hosting service.

        Version Changed:
            8.0:
            Made arguments keyword-only.

        Args:
            username (str):
                The username for the account.

            password (str):
                The password for the account.

            hosting_url (str, optional):
                The hosting URL for the service, if self-hosted.

            credentials (dict, optional):
                All credentials provided by the authentication form.

                This will contain the username, password, and anything else
                provided by that form.

            two_factor_auth_code (str, optional):
                The two-factor authentication code provided by the user.

            local_site_name (str, optional):
                The Local Site name, if any, that the account should be
                bound to.

            **kwargs (dict):
                Extra keyword arguments containing values from the
                repository's configuration.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                The credentials provided were not valid.

            reviewboard.hostingsvcs.errors.TwoFactorAuthCodeRequiredError:
                A two-factor authentication code is required to authorize
                this account. The request must be retried with the same
                credentials and with the ``two_factor_auth_code`` parameter
                provided.
        """
        raise NotImplementedError

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def check_repository(
        self,
        *,
        path: str | None,
        username: str | None,
        password: str | None,
        scmtool_class: type[SCMTool],
        local_site_name: str | None,
        **kwargs,
    ) -> None:
        """Check the validity of a repository configuration.

        This performs a check against the hosting service or repository
        to ensure that the information provided by the user represents
        a valid repository.

        This is passed in the repository details, such as the path and
        raw credentials, as well as the SCMTool class being used, the
        LocalSite's name (if any), and all field data from the
        HostingServiceForm as keyword arguments.

        Version Changed:
            8.0:
            Made arguments keyword-only.

        Args:
            path (str):
                The repository URL.

            username (str):
                The username to use.

            password (str):
                The password to use.

            scmtool_class (type):
                The subclass of :py:class:`~reviewboard.scmtools.core.SCMTool`
                that should be used.

            local_site_name (str):
                The name of the local site associated with the repository, or
                ``None``.

            **kwargs (dict):
                Additional keyword arguments, unique to each hosting service.

        Raises:
            reviewboard.hostingsvcs.errors.RepositoryError:
                The repository is not valid.
        """
        assert path is not None
        scmtool_class.check_repository(path, username, password,
                                       local_site_name)

    def get_file(
        self,
        repository: Repository,
        path: str,
        revision: str,
        *args,
        **kwargs,
    ) -> bytes:
        """Return the requested file.

        Files can only be returned from hosting services that support
        repositories.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve the file from.

            path (str):
                The file path.

            revision (str):
                The revision the file should be retrieved from.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Additional keyword arguments to pass to the SCMTool.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            NotImplementedError:
                If this hosting service does not support repositories.
        """
        if not self.supports_repositories:
            raise NotImplementedError

        return repository.get_scmtool().get_file(path, revision, **kwargs)

    def get_file_exists(
        self,
        repository: Repository,
        path: str,
        revision: str,
        *args,
        **kwargs,
    ) -> bool:
        """Return whether or not the given path exists in the repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to check for file existence.

            path (str):
                The file path.

            revision (str):
                The revision to check for file existence.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Additional keyword arguments to be passed to the SCMTool.

        Returns:
            bool:
            Whether or not the file exists at the given revision in the
            repository.

        Raises:
            NotImplementedError:
                If this hosting service does not support repositories.
        """
        if not self.supports_repositories:
            raise NotImplementedError

        return repository.get_scmtool().file_exists(path, revision, **kwargs)

    def get_branches(
        self,
        repository: Repository,
    ) -> Sequence[Branch]:
        """Return a list of all branches in the repositories.

        This should be implemented by subclasses, and is expected to return a
        list of Branch objects. One (and only one) of those objects should have
        the "default" field set to True.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository for which branches should be returned.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The branches.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error fetching branches.
        """
        raise NotImplementedError

    def get_commits(
        self,
        repository: Repository,
        branch: (str | None) = None,
        start: (str | None) = None,
    ) -> Sequence[Commit]:
        """Return a list of commits backward in history from a given point.

        This should be implemented by subclasses, and is expected to return a
        list of Commit objects (usually 30, but this is flexible depending on
        the limitations of the APIs provided.

        This can be called multiple times in succession using the "parent"
        field of the last entry as the start parameter in order to paginate
        through the history of commits in the repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve commits from.

            branch (str, optional):
                The branch to retrieve from.

                If this is not provided, the default branch will be used.

            start (str, optional):
                An optional starting revision.

                If this is not provided, the most recent commits will be
                returned.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The retrieved commits.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error fetching commits.
        """
        raise NotImplementedError

    def get_change(
        self,
        repository: Repository,
        revision: str,
    ) -> Commit:
        """Return an individual change.

        This method should be implemented by subclasses.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to get the change from.

            revision (str):
                The revision to retrieve.

        Returns:
            reviewboard.scmtools.core.Commit:
            The change.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error fetching the commit.
        """
        raise NotImplementedError

    def get_remote_repositories(
        self,
        owner: (str | None) = None,
        owner_type: (str | None) = None,
        filter_type: (str | None) = None,
        start: (int | None) = None,
        per_page: (int | None) = None,
        **kwargs,
    ) -> BasePaginator[RemoteRepository, Any]:
        """Return a list of remote repositories for the owner.

        This method should be implemented by subclasses.

        Args:
            owner (str, optional):
                The owner of the repositories.

                This is usually a username.

            owner_type (str, optional):
                A hosting service-specific indicator of what the owner is (such
                as a user or a group).

            filter_type (str, optional):
                Some hosting service-specific criteria to filter by.

            start (int, optional):
                The index to start at.

            per_page (int, optional):
                The number of results per page.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            reviewboard.hostingsvcs.utils.BasePaginator:
            A paginator for the returned repositories.
        """
        raise NotImplementedError

    def get_remote_repository(
        self,
        repository_id: str,
    ) -> RemoteRepository:
        """Return the remote repository for the ID.

        This method should be implemented by subclasses.

        Args:
            repository_id (str):
                The repository's identifier.

                This is unique to each hosting service.

        Returns:
            reviewboard.hostingsvcs.repository.RemoteRepository:
            The remote repository.

        Raises:
            django.core.excptions.ObjectDoesNotExist:
                If the remote repository does not exist.
        """
        raise NotImplementedError

    def normalize_patch(
        self,
        repository: Repository,
        patch: bytes,
        filename: str,
        revision: str,
    ) -> bytes:
        """Normalize a diff/patch file before it's applied.

        This can be used to take an uploaded diff file and modify it so that
        it can be properly applied. This may, for instance, uncollapse
        keywords or remove metadata that would confuse :command:`patch`.

        By default, this passes along the normalization to the repository's
        :py:class:`~reviewboard.scmtools.core.SCMTool`.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the patch is meant to apply to.

            patch (bytes):
                The diff/patch file to normalize.

            filename (str):
                The name of the file being changed in the diff.

            revision (str):
                The revision of the file being changed in the diff.

        Returns:
            bytes:
            The resulting diff/patch file.
        """
        return repository.get_scmtool().normalize_patch(patch=patch,
                                                        filename=filename,
                                                        revision=revision)

    @classmethod
    def get_repository_fields(
        cls,
        username: str,
        hosting_url: str | None,
        plan: str | None,
        tool_name: str,
        field_vars: dict[str, str],
    ) -> RepositoryFields:
        """Return the repository fields based on the given plan and tool.

        If the ``plan`` argument is specified, that will be used to fill in
        some tool-specific field values. Otherwise they will be retrieved from
        the :py:class:`HostingService`'s defaults.

        Args:
            username (str):
                The username.

            hosting_url (str):
                The URL of the repository.

            plan (str):
                The name of the plan.

            tool_name (str):
                The :py:attr:`~reviewboard.scmtools.core.SCMTool.name` of the
                :py:class:`~reviewboard.scmtools.core.SCMTool`.

            field_vars (dict):
                The field values from the hosting service form.

        Returns:
            dict:
            The filled in repository fields.

        Raises:
            KeyError:
               The provided plan is not valid for the hosting service.
        """
        if not cls.supports_repositories:
            raise NotImplementedError

        # Grab the list of fields for population below. We have to do this
        # differently depending on whether or not this hosting service has
        # different repository plans.
        fields: RepositoryFields = cls.get_field(plan, 'repository_fields')

        new_vars = field_vars.copy()
        new_vars['hosting_account_username'] = username

        if cls.self_hosted:
            assert hosting_url is not None

            new_vars['hosting_url'] = hosting_url
            new_vars['hosting_domain'] = urlparse(hosting_url)[1]

        results: RepositoryFields = {}

        assert tool_name in fields

        for field, value in fields[tool_name].items():
            try:
                results[field] = value % new_vars
            except KeyError as e:
                logger.exception('Failed to generate %s field for hosting '
                                 'service %s using %s and %r: Missing key %s',
                                 field, cls.name, value, new_vars, e)
                raise KeyError(
                    _('Internal error when generating %(field)s field '
                      '(Missing key "%(key)s"). Please report this.') % {
                        'field': field,
                        'key': e,
                    })

        return results

    def get_repository_hook_instructions(
        self,
        request: HttpRequest,
        repository: Repository,
    ) -> SafeString:
        """Return instructions for setting up incoming webhooks.

        Subclasses can override this (and set
        `has_repository_hook_instructions = True` on the subclass) to provide
        instructions that administrators can see when trying to configure an
        incoming webhook for the hosting service.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            repository (reviewboard.scmtools.models.Repository):
                The repository for webhook setup instructions.

        Returns:
            django.utils.text.SafeString:
            Rendered and escaped HTML for displaying to the user.
        """
        raise NotImplementedError

    @classmethod
    def get_bug_tracker_requires_username(
        cls,
        plan: (str | None) = None,
    ) -> bool:
        """Return whether or not the bug tracker requires usernames.

        Args:
            plan (str, optional):
                The name of the plan associated with the account.

        Returns:
            bool:
            ``True`` if a username is required. ``False`` if it is not.

        Raises:
            NotImplementedError:
                If the hosting service does not support bug tracking.
        """
        if not cls.supports_bug_trackers:
            raise NotImplementedError

        return (
            plan is not None and
            ('%(hosting_account_username)s' in
             cls.get_field(plan, 'bug_tracker_field', ''))
        )

    @classmethod
    def get_bug_tracker_field(
        cls,
        plan: str | None,
        field_vars: dict[str, str],
    ) -> str:
        """Return the bug tracker field for the given plan.

        Args:
            plan (str):
                The name of the plan associated with the account.

            field_vars (dict):
                The field values from the hosting service form.

        Returns:
            str:
            The value of the bug tracker field.

        Raises:
            KeyError:
               The provided plan is not valid for the hosting service.
        """
        if not cls.supports_bug_trackers:
            raise NotImplementedError

        bug_tracker_field = cls.get_field(plan, 'bug_tracker_field')

        if not bug_tracker_field:
            return ''

        try:
            return bug_tracker_field % field_vars
        except KeyError as e:
            logger.exception('Failed to generate %s field for hosting '
                             'service %s using %r: Missing key %s',
                             bug_tracker_field, cls.name, field_vars, e)
            raise KeyError(
                _('Internal error when generating %(field)s field '
                  '(Missing key "%(key)s"). Please report this.') % {
                    'field': bug_tracker_field,
                    'key': e,
                })

    @classmethod
    def get_field(
        cls,
        plan: str | None,
        name: str,
        default: (str | None) = None,
    ) -> Any:
        """Return the value of the field for the given plan.

        If ``plan`` is not ``None``, the field will be looked up in the plan
        configuration for the service. Otherwise the hosting service's default
        value will be used.

        Args:
            plan (str):
                The plan name.

            name (str):
                The field name.

            default (str, optional):
                A default value if the field is not present.

        Returns:
            str:
            The field value.
        """
        if cls.plans:
            assert plan

            for plan_name, info in cls.plans:
                if plan_name == plan and name in info:
                    return info[name]

        return getattr(cls, name, default)



"""Base class for a Review UI."""

from __future__ import annotations

import json
import logging
import os
from inspect import signature
from typing import (Any, ClassVar, Dict, Generic, Iterator, List, Optional,
                    Sequence, TYPE_CHECKING, TypeVar)
from urllib.parse import urlencode
from uuid import uuid4

import mimeparse
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse
from django.template.loader import render_to_string
from django.utils.translation import gettext as _
from housekeeping import func_deprecated
from typing_extensions import TypeAlias, TypedDict

from reviewboard.attachments.mimetypes import MIMETYPE_EXTENSIONS, score_match
from reviewboard.attachments.models import (FileAttachment,
                                            get_latest_file_attachments)
from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.diffviewer.models import FileDiff
from reviewboard.reviews.context import (make_review_request_context,
                                         should_view_draft)
from reviewboard.reviews.markdown_utils import (markdown_render_conditional,
                                                normalize_text_for_edit)
from reviewboard.reviews.models import (BaseComment,
                                        FileAttachmentComment,
                                        Review)
from reviewboard.reviews.models.review_request import FileAttachmentState
from reviewboard.site.urlresolvers import local_site_reverse


if TYPE_CHECKING:
    from django.http import HttpRequest
    from django.utils.safestring import SafeText
    from typelets.json import JSONDict

    from reviewboard.reviews.models import (
        ReviewRequest,
        ReviewRequestDraft,
    )
    from reviewboard.reviews.models.base_review_request_details import \
        BaseReviewRequestDetails


logger = logging.getLogger(__name__)


class SerializedCommentUser(TypedDict):
    """Serialized user for comment data.

    This must be kept in sync with the definitions in
    :file:`reviewboard/static/rb/js/reviews/models/commentData.ts`.

    Version Added:
        7.0
    """

    #: The user's full name, if available.
    name: str

    #: The user's username.
    username: str


class SerializedComment(TypedDict):
    """Serialized comment data to pass through to JavaScript.

    This must be kept in sync with the definitions in
    :file:`reviewboard/static/rb/js/reviews/models/commentData.ts`.

    Version Added:
        7.0
    """

    #: The ID of the comment.
    comment_id: int

    #: The rendered HTML version of the comment text.
    html: str

    #: Whether the comment opens an issue.
    issue_opened: bool

    #: The status of the issue, if one was opened.
    issue_status: str

    #: Whether the comment is part of the user's current draft review.
    localdraft: bool

    #: The ID of the comment that this comment is a reply to.
    reply_to_id: int

    #: The ID of the review that this comment is a part of.
    review_id: int

    #: The ID of the review request that this comment is on.
    review_request_id: int

    #: Whether the comment text should be rendered in Markdown.
    rich_text: bool

    #: The raw text of the comment.
    text: str

    #: The URL to link to for the comment.
    url: str

    #: Information about the author of the comment.
    user: SerializedCommentUser


#: A generic type for the reviewable object for a Review UI.
#:
#: Version Added:
#:     7.0
ReviewableType = TypeVar('ReviewableType')


#: A generic type for the type of comment used by a Review UI.
#:
#: Version Added:
#:     7.0
CommentType = TypeVar('CommentType', bound=BaseComment)


#: A generic type for the serialized comment data for a ReviewUI.
#:
#: Version Added:
#:     7.0
SerializedCommentType = TypeVar('SerializedCommentType',
                                bound=SerializedComment)


#: A type for the serialized comments for a review UI.
#:
#: Version Added:
#:     7.0
SerializedCommentBlocks: TypeAlias = Dict[str, List[SerializedCommentType]]


class ReviewUI(Generic[
    ReviewableType,
    CommentType,
    SerializedCommentType,
]):
    """Base class for a Review UI.

    Review UIs are interfaces for reviewing content of some type. They take a
    type of object and render a page around it, optionally allowing for the
    display of a diff view for the content. They can render context for
    comments made on the object, provide details for social media sharing (such
    as on a chat or social network).

    A Review UI makes use of a JavaScript side for the interaction, defined
    using :py:attr:`js_model_class` and :py:attr:`js_view_class`. The
    JavaScript side should interface with the API to create/update reviews and
    comments for the object being reviewed.
    """

    #: The display name for the Review UI.
    name: ClassVar[str] = 'Unknown file type'

    #: The template that renders the Review UI.
    #:
    #: Generally, subclasses should use the default template and render the
    #: UI using JavaScript.
    template_name: ClassVar[str] = 'reviews/ui/default.html'

    #: Whether the Review UI can be rendered inline in diffs and other places.
    #:
    #: If set, the Review UI will be able to be displayed within the diff
    #: viewer (and potentially other locations).
    allow_inline: ClassVar[bool] = False

    #: Whether this Review UI supports diffing two objects.
    supports_diffing: ClassVar[bool] = False

    #: A list of CSS bundle names to include on the Review UI's page.
    css_bundle_names: ClassVar[list[str]] = []

    #: A list of JavaScript bundle names to include on the Review UI's page.
    js_bundle_names: ClassVar[list[str]] = []

    #: A list of specific JavaScript URLs to include on the page.
    #:
    #: It is recommended that :py:attr:`js_bundle_names` be used instead
    #: where possible.
    js_files: ClassVar[list[str]] = []

    #: The list of MIME types that this Review UI supports.
    supported_mimetypes: ClassVar[list[str]] = []

    #: Whether this Review UI supports reviewing FileAttachment objects.
    supports_file_attachments: ClassVar[bool] = False

    #: Whether there's a file type mismatch when showing diffs.
    diff_type_mismatch: ClassVar[bool] = False

    #: Whether to load the Review UI's static media when rendered inline.
    #:
    #: If the extension managing the Review UI already loads its static media
    #: in the diff viewer page, then this should be set to ``False`` to
    #: prevent duplicate loading. This is useful for ensuring that state set
    #: by the static media gets shared amongst the extension instance and
    #: any review UIs on the page instead of being overwritten.
    #:
    #: Version Added:
    #:     8.0
    load_static_media_inline: ClassVar[bool] = True

    ######################
    # Instance variables #
    ######################

    #: The object being reviewed.
    obj: ReviewableType

    #: The object being compared against, if present.
    diff_against_obj: Optional[ReviewableType]

    #: The current HTTP request.
    #:
    #: This is only set once :py:meth:`render_to_string` is called.
    request: Optional[HttpRequest]

    #: Whether the Review UI is being rendered inline in the current request.
    #:
    #: Version Added:
    #:     8.0
    _inline: bool | None

    @property
    def js_model_class(self) -> str:
        """The name of the JavaScript model class to use for the Review UI.

        Type:
            str
        """
        if isinstance(self.obj, FileAttachment):
            return 'RB.DummyReviewable'
        else:
            return 'RB.AbstractReviewable'

    @property
    def js_view_class(self) -> str:
        """The name of the JavaScript view class to use for the Review UI.

        Type:
            str
        """
        if isinstance(self.obj, FileAttachment):
            return 'RB.DummyReviewableView'
        else:
            return 'RB.AbstractReviewableView'

    @property
    def object_key(self) -> str:
        """The key passed to the template representing the object.

        Type:
            str
        """
        if isinstance(self.obj, FileAttachment):
            return 'file'
        else:
            return 'obj'

    @property
    def diff_object_key(self) -> str:
        """The key passed to the template for an object to diff against.

        Type:
            str
        """
        if isinstance(self.obj, FileAttachment):
            return 'diff_against_file'
        else:
            return 'diff_against_obj'

    def __init__(
        self,
        review_request: ReviewRequest,
        obj: ReviewableType,
    ) -> None:
        """Initialize the Review UI.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request containing the object to review.

            obj (object):
                The object being reviewed.
        """
        self.review_request = review_request
        self.obj = obj
        self.diff_against_obj = None
        self.request = None
        self._inline = None

    def set_diff_against(
        self,
        obj: ReviewableType,
    ) -> None:
        """Set the object to generate a diff against.

        This can only be called on Review UIs that support diffing,
        and must be called before rendering.

        Args:
            obj (object):
                The object being diffed against.
        """
        assert self.supports_diffing

        self.diff_against_obj = obj

    def is_enabled_for(
        self,
        user: Optional[User] = None,
        review_request: Optional[ReviewRequest] = None,
        obj: Optional[object] = None,
        **kwargs,
    ) -> bool:
        """Return whether the Review UI is enabled under the given criteria.

        This can enable or disable a Review UI's functionality depending on
        the user, review request, or some state associated with one or more of
        those.

        When this is called, the arguments are always passed as keyword
        arguments. Subclasses don't need to accept all the arguments, as
        long as they take a ``**kwargs``.

        Args:
            user (django.contrib.auth.models.User, optional):
                The user to check.

            review_request (reviewboard.reviews.models.ReviewRequest,
                            optional):
                The review request to check.

            obj (object, optional):
                The object being reviewed.

            **kwargs (dict):
                Additional keyword arguments, for future expansion.

        Returns:
            bool:
            ``True`` if the Review UI is enabled for the given criteria.
            ``False`` otherwise.

            By default, Review UIs are always enabled.
        """
        if 'file_attachment' in kwargs:
            RemovedInReviewBoard90Warning.warn(
                'The file_attachment argument to ReviewUI.is_enabled_for is '
                'deprecated. Please pass obj= instead.')

        return True

    def render_to_response(
        self,
        request: HttpRequest,
    ) -> HttpResponse:
        """Render the Review UI to a response.

        This is used to render a page dedicated to the Review UI, complete
        with the standard Review Board chrome.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            django.http.HttpResponse:
            The HTTP response containing the page for the Review UI.
        """
        return HttpResponse(self.render_to_string(
            request=request,
            inline=request.GET.get('inline', '0') in ('1', 'true')))

    def render_to_string(
        self,
        request: HttpRequest,
        inline: bool = True,
    ) -> SafeText:
        """Render the Review UI to an HTML string.

        This renders the Review UI to a string for use in embedding into
        either an existing page or a new page.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            inline (bool, optional):
                Whether to render this such that it can be embedded into an
                existing page, instead of as a standalone page.

        Returns:
            django.utils.safestring.SafeText:
            The HTML for the Review UI.
        """
        self.request = request
        self._inline = inline

        try:
            context = self.build_render_context(request, inline=inline)

            return render_to_string(
                template_name=self.template_name,
                context=context,
                request=request)
        except Exception as e:
            logger.exception('Error when rendering %r: %s', self, e,
                             extra={'request': request})
            raise

    def build_render_context(
        self,
        request: HttpRequest,
        inline: bool = False,
        **kwargs,
    ) -> dict[str, Any]:
        """Build context for rendering the page.

        This computes the standard template context to use when rendering the
        page. Generally, subclasses should override
        :py:meth:`get_extra_context`, instead of this.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            inline (bool, optional):
                Whether to render this such that it can be embedded into an
                existing page, instead of as a standalone page.

            **kwargs (dict, unused):
                Keyword arguments, for future expansion.

        Returns:
            dict:
            The context to use in the template.
        """
        self.request = request
        self._inline = inline

        review_request = self.review_request
        last_activity_time = \
            review_request.get_last_activity_info()['timestamp']

        draft = review_request.get_draft(request.user)

        if should_view_draft(request=request, review_request=review_request,
                             draft=draft):
            review_request_details = draft or review_request
            caption = self.get_caption(draft)
        else:
            review_request_details = review_request
            caption = self.get_caption(None)

        context: dict[str, Any] = {
            'caption': caption,
            'comments': self.get_comments(),
            'last_activity_time': last_activity_time,
            'skip_static_media':
                self.request.GET.get('skip-static-media') == '1',
            'review_ui': self,
            'review_ui_uuid': str(uuid4()),
            self.object_key: self.obj,
            self.diff_object_key: self.diff_against_obj,
        }

        if inline:
            context.update({
                'base_template': 'reviews/ui/base_inline.html',
                'review_ui_inline': True,
            })

            if not self.load_static_media_inline:
                context['skip_static_media'] = True
        else:
            context.update({
                'base_template': 'reviews/ui/base.html',
                'review': review_request.get_pending_review(request.user),
                'review_ui_inline': False,
            })

        try:
            context.update(self.get_extra_context(request))
        except Exception as e:
            logger.exception('Error when calling get_extra_context for '
                             '%r: %s',
                             self, e,
                             extra={'request': request})
            raise

        is_file_attachment = isinstance(self.obj, FileAttachment)

        if is_file_attachment:
            social_page_title = (
                f'Attachment for Review Request '
                f'#{review_request.display_id}: {caption}'
            )

            state = review_request.get_file_attachment_state(self.obj)

            force_view_user_draft = (
                request.user != review_request.submitter and
                state in (FileAttachmentState.NEW_REVISION,
                          FileAttachmentState.NEW))
        else:
            social_page_title = (
                f'Reviewable for Review Request '
                f'#{review_request.display_id}: {caption}')
            force_view_user_draft = False

        context.update(make_review_request_context(
            request=request,
            review_request=review_request,
            draft=draft,
            force_view_user_draft=force_view_user_draft,
            review_request_details=review_request_details,
            social_page_image_url=self.get_page_cover_image_url(),
            social_page_title=social_page_title))

        if is_file_attachment and not inline:
            context['tabs'].append({
                'url': request.path,
                'text': _('File'),
            })

            prev_file_attachment, next_file_attachment = \
                self._get_adjacent_file_attachments(review_request_details)

            context.update({
                'next_file_attachment': next_file_attachment,
                'prev_file_attachment': prev_file_attachment,
            })

        return context

    def get_page_cover_image_url(self) -> Optional[str]:
        """Return the URL to an image used to depict this on other sites.

        The returned image URL will be used for services like Facebook, Slack,
        Twitter, etc. when linking to the reviewable object. This may be
        anything from a standard thumbnail to a full-size image.

        By default, no image URL is returned.

        Returns:
            str:
            The absolute URL to an image used to depict the reviewable object.
        """
        return None

    def get_comments(self) -> Sequence[CommentType]:
        """Return all existing comments on the reviewable object.

        Subclasses must override this.

        Returns:
            list of reviewboard.reviews.models.base_comment.BaseComment:
            The list of comments for the page.
        """
        if isinstance(self.obj, FileAttachment):
            comments = FileAttachmentComment.objects.filter(
                file_attachment_id=self.obj.pk)

            if self.diff_against_obj:
                assert isinstance(self.diff_against_obj, FileAttachment)
                comments = comments.filter(
                    diff_against_file_attachment_id=self.diff_against_obj.pk)
            else:
                comments = comments.filter(
                    diff_against_file_attachment_id__isnull=True)

            return list(comments)

        raise NotImplementedError

    def get_caption(
        self,
        draft: Optional[ReviewRequestDraft] = None,
    ) -> str:
        """Return the caption to show for the reviewable object.

        This defaults to requiring ``caption`` and ``draft_caption`` attributes
        on the reviewable object. Subclasses can override this to use something
        else.

        Args:
            draft (reviewboard.reviews.models.ReviewRequestDraft, optional):
                The active review request draft for the user, if any.

        Returns:
            str:
            The caption for the reviewable object.
        """
        if isinstance(self.obj, FileAttachment):
            if draft and self.obj.draft_caption:
                return self.obj.draft_caption
            else:
                return self.obj.caption

        raise NotImplementedError

    def get_comment_thumbnail(
        self,
        comment: CommentType,
    ) -> Optional[SafeText]:
        """Return an HTML thumbnail for a comment.

        If comment thumbnails are possible for the reviewable object, this
        function should return HTML for the thumbnail.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to return a thumbnail for.

        Returns:
            django.utils.safestring.SafeText:
            The HTML for a thumbnail for the comment, or ``None`` if one
            can't be generated (using the default thumbnailing for the
            comment type, if one exists).
        """
        return None

    def get_comment_link_url(
        self,
        comment: CommentType,
    ) -> str:
        """Return a URL for linking to a comment.

        Subclasses must override this.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to return a link for.

        Returns:
            str:
            The URL to link to the comment.
        """
        if isinstance(comment, FileAttachmentComment):
            assert isinstance(self.obj, FileAttachment)

            if self.obj.is_from_diff:
                return self._get_diff_attachment_comment_link_url(comment)
            else:
                return local_site_reverse(
                    'file-attachment',
                    local_site=self.review_request.local_site,
                    kwargs={
                        'review_request_id': self.review_request.display_id,
                        'file_attachment_id': self.obj.pk,
                    })

        raise NotImplementedError

    def get_comment_link_text(
        self,
        comment: CommentType,
    ) -> Optional[str]:
        """Return the text to link to a comment.

        This must be implemented by subclasses.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to return text for.

        Returns:
            str:
            The text used to link to the comment.
        """
        if isinstance(self.obj, FileAttachment):
            return self.obj.display_name

        raise NotImplementedError

    def get_extra_context(
        self,
        request: HttpRequest,
    ) -> dict[str, Any]:
        """Return extra context to use when rendering the Review UI.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            The context to provide to the template.
        """
        return {}

    def get_js_model_data(self) -> JSONDict:
        """Return data to pass to the JavaScript Model during instantiation.

        This data will be passed as attributes to the reviewable model
        when constructed.

        Returns:
            dict:
            The attributes to pass to the model.
        """
        data: dict[str, Any] = {}
        obj = self.obj

        if isinstance(obj, FileAttachment):
            review_request = self.review_request
            state = review_request.get_file_attachment_state(obj)
            data.update({
                'fileAttachmentID': obj.pk,
                'fileRevision': obj.attachment_revision,
                'filename': obj.orig_filename,
                'state': state.value,
            })

            if (inline := self._inline) is not None:
                data['renderedInline'] = inline

            if obj.attachment_history is not None:
                request = self.request
                include_draft = (
                    request is not None and should_view_draft(
                        request=request,
                        review_request=review_request,
                        draft=review_request.get_draft(request.user)))

                attachments = list(
                    FileAttachment.objects
                    .filter(attachment_history=obj.attachment_history)
                    .order_by('attachment_revision'))

                if attachments:
                    revisions = [
                        attachment.pk
                        for attachment in attachments[:-1]
                    ]
                    last_attachment = attachments[-1]

                    # If the last attachment is newly added in a draft, we only
                    # want to include it in the revision list if the draft is
                    # visible to the user.
                    if (include_draft or
                        review_request.get_file_attachment_state(
                            last_attachment,
                        ) not in [
                            FileAttachmentState.NEW,
                            FileAttachmentState.NEW_REVISION,
                        ]):
                        revisions.append(last_attachment.pk)

                    data['attachmentRevisionIDs'] = revisions
                    data['numRevisions'] = len(revisions)

            if self.diff_against_obj:
                diff_against_obj = self.diff_against_obj
                assert isinstance(diff_against_obj, FileAttachment)

                data['diffCaption'] = diff_against_obj.display_name
                data['diffAgainstFileAttachmentID'] = diff_against_obj.pk
                data['diffRevision'] = diff_against_obj.attachment_revision
                data['diffTypeMismatch'] = self.diff_type_mismatch

        return data

    def get_js_view_data(self) -> JSONDict:
        """Return data to pass to the JavaScript View during instantiation.

        This data will be passed as options to the reviewable view
        when constructed.

        Returns:
            dict:
            The options to pass to the view.
        """
        data: dict[str, Any] = {}

        if (inline := self._inline) is not None:
            data['renderedInline'] = inline

        return data

    def get_comments_json(self) -> str:
        """Return a JSON-serialized representation of comments for a template.

        The result of this can be used directly in a template to provide
        comments to JavaScript functions.

        Returns:
            str:
            Serialized JSON content representing the comments on the reviewable
            object.
        """
        try:
            return json.dumps(
                self.serialize_comments(self.get_comments()),
                sort_keys=True)
        except Exception as e:
            logger.exception('Error When calling serialize_comments for '
                             '%r: %s',
                             self, e,
                             extra={'request': self.request})
            raise

    def serialize_comments(
        self,
        comments: Sequence[CommentType],
    ) -> SerializedCommentBlocks[SerializedCommentType]:
        """Serialize the comments for the Review UI target.

        By default, this will return a "flat" array of comments, but it can be
        overridden in order to group comments by identifier or region.

        Args:
            comments (list of reviewboard.reviews.models.base_comment.
                      BaseComment):
                The list of objects to serialize. This will be the result of
                :py:meth:`get_comments`.

        Returns:
            SerializedCommentBlocks:
            The set of serialized comment data.
        """
        result: SerializedCommentBlocks[SerializedCommentType] = {}

        for i, comment in enumerate(self.flat_serialized_comments(comments)):
            result[str(i)] = [comment]

        return result

    def flat_serialized_comments(
        self,
        comments: Sequence[CommentType],
    ) -> Iterator[SerializedCommentType]:
        """Yield the serialized comments.

        This will go through the list of comments and filter out any which
        should not be shown (ones which are in other users' drafts), then yield
        the serialized form for each.

        Yields:
            SerializedComment:
            The serialized comment.
        """
        assert self.request is not None
        user = self.request.user

        for comment in comments:
            try:
                review = comment.get_review()
            except Review.DoesNotExist:
                logger.error('Missing Review for comment %r',
                             comment,
                             extra={'request': self.request})
                continue

            try:
                if review and (review.public or review.user_id == user.pk):
                    yield self.serialize_comment(comment)
            except Exception as e:
                logger.exception(
                    'Error when calling serialize_comment for %r: %s',
                    comment, e,
                    extra={'request': self.request})
                raise

    def serialize_comment(
        self,
        comment: CommentType,
    ) -> SerializedComment:
        """Serialize a comment.

        This will provide information on the comment that may be useful
        to the JavaScript code.

        Subclasses that want to add additional data should generally
        augment the result of this function and not replace it.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to serialize.

        Returns:
            SerializedComment:
            The serialized comment data.
        """
        review = comment.get_review()

        assert self.request is not None
        user = self.request.user

        data: SerializedComment = {
            'comment_id': comment.pk,
            'issue_opened': comment.issue_opened,
            'issue_status': comment.issue_status_to_string(
                comment.issue_status),
            'html': markdown_render_conditional(comment.text,
                                                comment.rich_text),
            'localdraft': review.user_id == user.pk and not review.public,
            'reply_to_id': comment.reply_to_id,
            'review_id': review.pk,
            'review_request_id': review.review_request_id,
            'rich_text': comment.rich_text,
            'text': normalize_text_for_edit(user, comment.text,
                                            comment.rich_text),
            'url': comment.get_review_url(),
            'user': {
                'username': review.user.username,
                'name': review.user.get_profile().get_display_name(user),
            },
        }

        if isinstance(comment, FileAttachmentComment):
            data.update(comment.extra_data)

        return data

    def _get_adjacent_file_attachments(
        self,
        review_request_details: BaseReviewRequestDetails,
    ) -> tuple[Optional[FileAttachment], Optional[FileAttachment]]:
        """Return the next and previous file attachments.

        The next and previous file attachments are the file attachments that
        occur before and after this one in the review request details view.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            tuple:
            A 2-tuple of the previous and next file attachments, which will
            either be ``None`` (if there isn't a previous or next file
            attachment) or
            :py:class:`~reviewboard.attachments.models.FileAttachment`
            instances.
        """
        assert isinstance(self.obj, FileAttachment)
        assert self.request is not None

        file_attachments = iter(get_latest_file_attachments(
            review_request_details.get_file_attachments()))

        prev_obj = None
        next_obj = None
        user = self.request.user

        for obj in file_attachments:
            if obj.pk == self.obj.pk:
                break

            if obj.is_review_ui_accessible_by(user):
                prev_obj = obj

        try:
            obj = next(file_attachments)

            if obj.is_review_ui_accessible_by(user):
                next_obj = obj
        except StopIteration:
            pass

        return prev_obj, next_obj

    def _get_diff_attachment_comment_link_url(
        self,
        comment: FileAttachmentComment,
    ) -> str:
        """Return the URL for linking to a comment on a diff file attachment.

        This will inspect a file attachment comment which was made on an
        attachment that's part of a diff, and return the URL to the diff viewer
        with the correct revisions.

        Args:
            comment (reviewboard.reviews.models.FileAttachmentComment):
                The file attachment comment.

        Returns:
            str:
            The link to the relevant diffviewer revision that the comment was
            made on.
        """
        revision_info = comment.get_comment_diff_revision_info()
        assert revision_info is not None

        diff_revision = revision_info['diff_revision']
        interdiff_revision = revision_info['interdiff_revision']
        base_commit_id = revision_info['base_commit_id']
        tip_commit_id = revision_info['tip_commit_id']
        modified_filediff = revision_info['modified_filediff']
        modified_diffset = revision_info['modified_diffset']

        filediff_id: int = modified_filediff.pk
        query: dict[str, Any] = {}

        if base_commit_id:
            query['base-commit-id'] = base_commit_id

        if tip_commit_id:
            query['tip-commit-id'] = tip_commit_id

        if (modified_diffset.commit_count and
            not base_commit_id and
            not tip_commit_id):
            # For review requests created with commit history, we can have
            # multiple FileDiffs for the same file--one as part of the commit,
            # and one as part of the cumulative diff. The file attachment for a
            # binary file gets uploaded to the FileDiff corresponding to the
            # commit.
            #
            # When the user loads the cumulative diff, we'll find the file
            # attachments that correspond to those files. See
            # :py:meth:`reviewboard.attachments.managers.
            # FileAttachmentManager.get_for_filediff`.
            #
            # When we've reached this point in the code, we know that the user
            # made their comment on the cumulative diff (we don't have any
            # base-commit-id or tip-commit-id parameters to stick in the URL,
            # yet the diffset has commits). We therefore need to do the reverse
            # of what ``get_for_filediff` does, and find the FileDiff in the
            # cumulative diff that corresponds to the commit-specific one we
            # have.
            filediff_id = (
                FileDiff.objects
                .filter(
                    diffset_id=modified_diffset.pk,
                    commit__isnull=True,
                    dest_file=modified_filediff.dest_file)
                .values_list('pk', flat=True)
            )[0]

        review_request = self.review_request

        if interdiff_revision:
            url = local_site_reverse(
                'view-interdiff',
                local_site=review_request.local_site,
                kwargs={
                    'interdiff_revision': interdiff_revision,
                    'review_request_id': review_request.display_id,
                    'revision': diff_revision,
                })
        else:
            url = local_site_reverse(
                'view-diff-revision',
                local_site=review_request.local_site,
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diff_revision,
                })

        if query:
            url += f'?{urlencode(query)}'

        return f'{url}#file{filediff_id}'

    @classmethod
    def for_object(
        cls,
        obj: object,
    ) -> Optional[type[ReviewUI]]:
        """Return the Review UI that is the best fit for a given object.

        Args:
            obj (object):
                The object to review.

        Returns:
            class:
            The Review UI class for the given object, or ``None`` if a
            suitable one could not be found.
        """
        is_file_attachment = isinstance(obj, FileAttachment)
        mime_str: Optional[str] = None
        extension: Optional[str] = None

        if is_file_attachment:
            mime_str = obj.mimetype

            assert obj.filename is not None
            extension = os.path.splitext(obj.filename)[1]

        if not mime_str:
            return None

        if extension and extension in MIMETYPE_EXTENSIONS:
            mimetype = MIMETYPE_EXTENSIONS[extension]
        else:
            try:
                mimetype = mimeparse.parse_mime_type(mime_str)
            except Exception as e:
                logger.error('Unable to parse MIME type "%s" for '
                             'reviewable object %s: %s',
                             mime_str, obj, e)
                return None

        best_score = 0
        best_fit: Optional[type[ReviewUI]] = None

        from reviewboard.reviews.ui import review_ui_registry

        for review_ui in review_ui_registry:
            if not (is_file_attachment and
                    review_ui.supports_file_attachments):
                continue

            for mt in review_ui.supported_mimetypes:
                try:
                    score = score_match(mimeparse.parse_mime_type(mt),
                                        mimetype)

                    if score > best_score:
                        best_score = score
                        best_fit = review_ui
                except ValueError:
                    continue

        return best_fit


class FileAttachmentReviewUI(ReviewUI):
    """Base class for Review UIs for file attachments.

    Review UIs that deal with
    :py:class:`~reviewboard.attachments.models.FileAttachment` objects can
    subclass this to provide the common functionality for their Review UI.

    This class handles fetching and serializing comments, locating a correct
    subclass for a given mimetype, and feeding data to the JavaScript
    :js:class:`RB.AbstractReviewable` model.

    This also handles much of the work for diffing file attachments.
    """

    supports_file_attachments = True

    @classmethod
    def get_best_handler(
        cls,
        mimetype: tuple[str, str, str],
    ) -> tuple[float, Optional[type[ReviewUI]]]:
        """Return the Review UI and score that that best fit the mimetype.

        Args:
            mimetype (tuple):
                A parsed mimetype to find the best review UI for. This is a
                3-tuple of the type, subtype, and parameters as returned by
                :py:func:`mimeparse.parse_mime_type`.

        Returns:
            tuple:
            A tuple of ``(best_score, review_ui)``, or ``(0, None)`` if one
            could not be found.
        """
        RemovedInReviewBoard90Warning.warn(
            'FileAttachmentReviewUI.get_best_handler is deprecated and will '
            'be removed in Review Board 9.')

        best_score = 0
        best_fit = None

        from reviewboard.reviews.ui import review_ui_registry

        for review_ui in review_ui_registry:
            for mt in review_ui.supported_mimetypes:
                try:
                    score = score_match(mimeparse.parse_mime_type(mt),
                                        mimetype)

                    if score > best_score:
                        best_score = score
                        best_fit = review_ui
                except ValueError:
                    continue

        return best_score, best_fit

    @classmethod
    def for_type(
        cls,
        attachment: FileAttachment,
    ) -> Optional[ReviewUI]:
        """Return the Review UI that is the best fit for a file attachment.

        Args:
            attachment (reviewboard.attachments.models.FileAttachments):
                The file attachment to locate a Review UI for.

        Returns:
            FileAttachmentReviewUI:
            The Review UI for the attachment, or ``None`` if a suitable one
            could not be found.
        """
        RemovedInReviewBoard90Warning.warn(
            'FileAttachmentReviewUI.for_type is deprecated and will '
            'be removed in Review Board 9. Callers should instead use '
            'ReviewUI.for_object')

        if attachment.mimetype:
            # Override the mimetype if mimeparse is known to misinterpret this
            # type of file as 'octet-stream'
            assert attachment.filename is not None
            extension = os.path.splitext(attachment.filename)[1]

            if extension in MIMETYPE_EXTENSIONS:
                mimetype = MIMETYPE_EXTENSIONS[extension]
            else:
                try:
                    mimetype = mimeparse.parse_mime_type(attachment.mimetype)
                except Exception:
                    logger.error('Unable to parse MIME type "%s" for %s',
                                 attachment.mimetype, attachment)
                    return None

            score, handler = cls.get_best_handler(mimetype)

            if handler:
                try:
                    return handler(attachment.get_review_request(), attachment)
                except ObjectDoesNotExist as e:
                    logger.error('Unable to load review UI for %s: %s',
                                 attachment, e)
                except Exception as e:
                    logger.exception('Error instantiating '
                                     'FileAttachmentReviewUI %r: %s',
                                     handler, e)

        return None


class DiffMismatchReviewUI(ReviewUI):
    """A special review UI to show when the file types in a diff don't match.
    """

    diff_type_mismatch = True
    supports_diffing = True


def register_ui(review_ui: type[ReviewUI]) -> None:
    """Register a Review UI class.

    This will register a Review UI. Review Board will use it to display a UI
    when reviewing a supported file attachment.

    Args:
        review_ui (type):
            The Review UI to register. This must be a subclass of
            :py:class:`ReviewUI`.

    Raises:
        TypeError:
            The provided Review UI class is not of a compatible type.
    """
    if not issubclass(review_ui, ReviewUI):
        raise TypeError('Only ReviewUI subclasses can be registered')

    from reviewboard.reviews.ui import review_ui_registry

    review_ui_registry.register(review_ui)


def unregister_ui(review_ui: type[ReviewUI]) -> None:
    """Unregister a Review UI class.

    This will unregister a previously registered Review UI.

    Only :py:class:`ReviewUI` subclasses are supported. The class must have
    been registered beforehand or a ValueError will be thrown.

    Args:
        review_ui (type):
            The Review UI to unregister. This must be a subclass of
            :py:class:`ReviewUI`, and must have been registered before.

    Raises:
        TypeError:
            The provided Review UI class is not of a compatible type.

        ValueError:
            The provided Review UI was not previously registered.
    """
    if not issubclass(review_ui, ReviewUI):
        raise TypeError('Only ReviewUI subclasses can be '
                        'unregistered')

    from reviewboard.reviews.ui import review_ui_registry

    review_ui_registry.unregister(review_ui)


@func_deprecated(RemovedInReviewBoard90Warning)
def is_review_ui_enabled_for(
    *,
    review_ui: ReviewUI[FileAttachment, FileAttachmentComment,
                        SerializedComment],
    request: HttpRequest,
    review_request: ReviewRequest,
    file_attachment: FileAttachment,
) -> bool:
    """Return whether a Review UI is enabled for the given object.

    Deprecated:
        7.0.3:
        Use :py:meth:`~reviewboard.attachments.models.FileAttachment.
        is_review_ui_accessible_by` instead. This function will be removed
        in 8.0.

    Version Added:
        7.0.2

    Args:
        review_ui (reviewboard.reviews.ui.base.ReviewUI):
            The Review UI to check.

        request (django.http.HttpRequest):
            The HTTP request from the client.

        review_request (reviewboard.reviews.models.ReviewRequest):
            The review request.

        file_attachment (reviewboard.attachments.models.FileAttachment):
            The file attachment.

    Returns:
        bool:
        ``True`` if the review UI is enabled for the given attachment.
    """
    user = request.user

    params = signature(review_ui.is_enabled_for).parameters

    try:
        if 'file_attachment' in params:
            RemovedInReviewBoard90Warning.warn(
                'The file_attachment parameter to ReviewUI.is_enabled_for '
                'has been removed. Please use obj= instead in Review UI %r'
                % review_ui)

            return review_ui.is_enabled_for(
                user=user,
                review_request=review_request,
                file_attachment=file_attachment)
        else:
            return review_ui.is_enabled_for(
                user=user,
                review_request=review_request,
                obj=file_attachment)
    except Exception as e:
        logger.exception(
            'Error when calling is_enabled_for for ReviewUI %r: %s',
            review_ui, e, extra={'request': request})

        return False



"""Code safety support.

These modules are used to help ensure code going up for review passes certain
safety checks.

Version Added:
    5.0
"""

from djblets.registries.importer import lazy_import_registry


code_safety_checker_registry = \
    lazy_import_registry('reviewboard.codesafety.checkers.registry',
                         'CodeSafetyCheckerRegistry')



"""Unit tests for reviewboard.diffviewer.models.filediff."""

from itertools import chain

from reviewboard.diffviewer.models import DiffSet, FileDiff
from reviewboard.diffviewer.tests.test_diffutils import \
    BaseFileDiffAncestorTests
from reviewboard.testing import TestCase


class FileDiffTests(TestCase):
    """Unit tests for FileDiff."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(FileDiffTests, self).setUp()

        diff = (
            b'diff --git a/README b/README\n'
            b'index 3d2b777..48272a3 100644\n'
            b'--- README\n'
            b'+++ README\n'
            b'@@ -2 +2,2 @@\n'
            b'-blah blah\n'
            b'+blah!\n'
            b'+blah!!\n'
        )

        self.repository = self.create_repository(tool_name='Test')
        self.diffset = DiffSet.objects.create(name='test',
                                              revision=1,
                                              repository=self.repository)
        self.filediff = FileDiff(source_file='README',
                                 dest_file='README',
                                 diffset=self.diffset,
                                 diff64=diff,
                                 parent_diff64=b'')

    def test_get_line_counts_with_defaults(self):
        """Testing FileDiff.get_line_counts with default values"""
        counts = self.filediff.get_line_counts()

        self.assertIn('raw_insert_count', counts)
        self.assertIn('raw_delete_count', counts)
        self.assertIn('insert_count', counts)
        self.assertIn('delete_count', counts)
        self.assertIn('replace_count', counts)
        self.assertIn('equal_count', counts)
        self.assertIn('total_line_count', counts)
        self.assertEqual(counts['raw_insert_count'], 2)
        self.assertEqual(counts['raw_delete_count'], 1)
        self.assertEqual(counts['insert_count'], 2)
        self.assertEqual(counts['delete_count'], 1)
        self.assertIsNone(counts['replace_count'])
        self.assertIsNone(counts['equal_count'])
        self.assertIsNone(counts['total_line_count'])

        diff_hash = self.filediff.diff_hash
        self.assertEqual(diff_hash.insert_count, 2)
        self.assertEqual(diff_hash.delete_count, 1)

    def test_set_line_counts(self):
        """Testing FileDiff.set_line_counts"""
        self.filediff.set_line_counts(
            raw_insert_count=1,
            raw_delete_count=2,
            insert_count=3,
            delete_count=4,
            replace_count=5,
            equal_count=6,
            total_line_count=7)

        counts = self.filediff.get_line_counts()
        self.assertEqual(counts['raw_insert_count'], 1)
        self.assertEqual(counts['raw_delete_count'], 2)
        self.assertEqual(counts['insert_count'], 3)
        self.assertEqual(counts['delete_count'], 4)
        self.assertEqual(counts['replace_count'], 5)
        self.assertEqual(counts['equal_count'], 6)
        self.assertEqual(counts['total_line_count'], 7)

        diff_hash = self.filediff.diff_hash
        self.assertEqual(diff_hash.insert_count, 1)
        self.assertEqual(diff_hash.delete_count, 2)

    def test_long_filenames(self):
        """Testing FileDiff with long filenames (1024 characters)"""
        long_filename = 'x' * 1024

        filediff = FileDiff.objects.create(source_file=long_filename,
                                           dest_file='foo',
                                           diffset=self.diffset)
        self.assertEqual(filediff.source_file, long_filename)

    def test_diff_hashes(self):
        """Testing FileDiff with multiple entries and same diff data
        deduplicates data
        """
        data = (
            b'diff -rcN orig_src/foo.c new_src/foo.c\n'
            b'*** orig_src/foo.c\t2007-01-24 02:11:31.000000000 -0800\n'
            b'--- new_src/foo.c\t2007-01-24 02:14:42.000000000 -0800\n'
            b'***************\n'
            b'*** 1,5 ****\n'
            b'  int\n'
            b'  main()\n'
            b'  {\n'
            b'! \tprintf("foo\n");\n'
            b'  }\n'
            b'--- 1,8 ----\n'
            b'+ #include <stdio.h>\n'
            b'+ \n'
            b'  int\n'
            b'  main()\n'
            b'  {\n'
            b'! \tprintf("foo bar\n");\n'
            b'! \treturn 0;\n'
            b'  }\n')

        filediff1 = FileDiff.objects.create(diff=data, diffset=self.diffset)
        filediff2 = FileDiff.objects.create(diff=data, diffset=self.diffset)

        self.assertEqual(filediff1.diff_hash, filediff2.diff_hash)

    def test_get_base_filediff(self):
        """Testing FileDiff.get_base_filediff"""
        commit1 = self.create_diffcommit(
            diffset=self.diffset,
            commit_id='r1',
            parent_id='r0',
            diff_contents=(
                b'diff --git a/ABC b/ABC\n'
                b'index 94bdd3e..197009f 100644\n'
                b'--- ABC\n'
                b'+++ ABC\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-line!\n'
                b'+line..\n'
            ))
        commit2 = self.create_diffcommit(
            diffset=self.diffset,
            commit_id='r2',
            parent_id='r1',
            diff_contents=(
                b'diff --git a/README b/README\n'
                b'index 94bdd3e..197009f 100644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hello, world!\n'
                b'+Hi, world!\n'
            ))
        commit3 = self.create_diffcommit(
            diffset=self.diffset,
            commit_id='r3',
            parent_id='r2',
            diff_contents=(
                b'diff --git a/FOO b/FOO\n'
                b'index 84bda3e..b975034 100644\n'
                b'--- FOO\n'
                b'+++ FOO\n'
                b'@@ -1,1 +0,0 @@\n'
                b'-Some line\n'
            ))
        commit4 = self.create_diffcommit(
            diffset=self.diffset,
            commit_id='r4',
            parent_id='r3',
            diff_contents=(
                b'diff --git a/README b/README\n'
                b'index 197009f..87abad9 100644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hi, world!\n'
                b'+Yo, world.\n'
            ))

        self.diffset.finalize_commit_series(
            cumulative_diff=(
                b'diff --git a/ABC b/ABC\n'
                b'index 94bdd3e..197009f 100644\n'
                b'--- ABC\n'
                b'+++ ABC\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-line!\n'
                b'+line..\n'
                b'diff --git a/FOO b/FOO\n'
                b'index 84bda3e..b975034 100644\n'
                b'--- FOO\n'
                b'+++ FOO\n'
                b'@@ -1,1 +0,0 @@\n'
                b'-Some line\n'
                b'diff --git a/README b/README\n'
                b'index 94bdd3e..87abad9 100644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hello, world!\n'
                b'+Yo, world.\n'
            ),
            validation_info=None,
            validate=False,
            save=True)

        filediff1 = commit1.files.get()
        filediff2 = commit2.files.get()
        filediff3 = commit3.files.get()
        filediff4 = commit4.files.get()

        for commit in (commit1, commit2, commit3, commit4):
            self.assertIsNone(filediff1.get_base_filediff(base_commit=commit))
            self.assertIsNone(filediff2.get_base_filediff(base_commit=commit))
            self.assertIsNone(filediff3.get_base_filediff(base_commit=commit))

        self.assertIsNone(filediff4.get_base_filediff(base_commit=commit1))
        self.assertEqual(filediff4.get_base_filediff(base_commit=commit2),
                         filediff2)
        self.assertEqual(filediff4.get_base_filediff(base_commit=commit3),
                         filediff2)
        self.assertEqual(filediff4.get_base_filediff(base_commit=commit4),
                         filediff2)

    def test_get_base_filediff_without_commit(self):
        """Testing FileDiff.get_base_filediff without associated commit"""
        filediff = self.create_filediff(self.diffset)

        self.assertIsNone(filediff.get_base_filediff(base_commit=None))

    def test_is_symlink_with_true(self):
        """Testing FileDiff.is_symlink with True"""
        filediff = self.create_filediff(self.diffset)
        filediff.is_symlink = True

        # Explicitly test against the booleans, to avoid truthiness tests.
        self.assertIs(filediff.is_symlink, True)
        self.assertIs(filediff.extra_data.get('is_symlink'), True)

    def test_is_symlink_with_false(self):
        """Testing FileDiff.is_symlink with False"""
        filediff = self.create_filediff(self.diffset)
        filediff.extra_data['is_symlink'] = True
        filediff.is_symlink = False

        # Explicitly test against the booleans, to avoid truthiness tests.
        self.assertIs(filediff.is_symlink, False)
        self.assertIs(filediff.extra_data.get('is_symlink'), False)

    def test_old_symlink_target(self):
        """Testing FileDiff.old_symlink_target"""
        filediff = self.create_filediff(self.diffset)
        filediff.old_symlink_target = 'old/path'

        self.assertEqual(filediff.old_symlink_target, 'old/path')
        self.assertEqual(filediff.extra_data.get('old_symlink_target'),
                         'old/path')

    def test_new_symlink_target(self):
        """Testing FileDiff.new_symlink_target"""
        filediff = self.create_filediff(self.diffset)
        filediff.new_symlink_target = 'new/path'

        self.assertEqual(filediff.new_symlink_target, 'new/path')
        self.assertEqual(filediff.extra_data.get('new_symlink_target'),
                         'new/path')

    def test_old_unix_mode(self):
        """Testing FileDiff.old_unix_mode"""
        filediff = self.create_filediff(self.diffset)
        filediff.old_unix_mode = '0100644'

        self.assertEqual(filediff.old_unix_mode, '0100644')
        self.assertEqual(filediff.extra_data.get('old_unix_mode'), '0100644')

    def test_new_unix_mode(self):
        """Testing FileDiff.new_unix_mode"""
        filediff = self.create_filediff(self.diffset)
        filediff.new_unix_mode = '0100750'

        self.assertEqual(filediff.new_unix_mode, '0100750')
        self.assertEqual(filediff.extra_data.get('new_unix_mode'), '0100750')


class FileDiffAncestorTests(BaseFileDiffAncestorTests):
    """Unit tests for FileDiff.get_ancestors"""

    def setUp(self):
        super(FileDiffAncestorTests, self).setUp()

        self.set_up_filediffs()

    def test_get_ancestors_minimal(self):
        """Testing FileDiff.get_ancestors with minimal=True"""
        ancestors = {}

        with self.assertNumQueries(9):
            for filediff in self.filediffs:
                ancestors[filediff] = filediff.get_ancestors(
                    minimal=True,
                    filediffs=self.filediffs)

        self._check_ancestors(ancestors, minimal=True)

    def test_get_ancestors_full(self):
        """Testing FileDiff.get_ancestors with minimal=False"""
        ancestors = {}

        with self.assertNumQueries(len(self.filediffs)):
            for filediff in self.filediffs:
                ancestors[filediff] = filediff.get_ancestors(
                    minimal=False,
                    filediffs=self.filediffs)

        self._check_ancestors(ancestors, minimal=False)

    def test_get_ancestors_cached(self):
        """Testing FileDiff.get_ancestors with cached results"""
        ancestors = {}

        for filediff in self.filediffs:
            filediff.get_ancestors(minimal=True, filediffs=self.filediffs)

        for filediff in self.filediffs:
            with self.assertNumQueries(0):
                ancestors[filediff] = filediff.get_ancestors(
                    minimal=True,
                    filediffs=self.filediffs)

        self._check_ancestors(ancestors, minimal=True)

    def test_get_ancestors_no_update(self):
        """Testing FileDiff.get_ancestors without caching"""
        ancestors = {}

        for filediff in self.filediffs:
            with self.assertNumQueries(0):
                ancestors[filediff] = filediff.get_ancestors(
                    minimal=True,
                    filediffs=self.filediffs,
                    update=False)

        self._check_ancestors(ancestors, minimal=True)

    def test_get_ancestors_no_filediffs(self):
        """Testing FileDiff.get_ancestors when no FileDiffs are provided"""
        ancestors = {}

        with self.assertNumQueries(2 * len(self.filediffs)):
            for filediff in self.filediffs:
                ancestors[filediff] = filediff.get_ancestors(minimal=True)

        self._check_ancestors(ancestors, minimal=True)

    def test_get_ancestors_cached_no_filediffs(self):
        """Testing FileDiff.get_ancestors with cached results when no
        FileDiffs are provided
        """
        ancestors = {}

        for filediff in self.filediffs:
            filediff.get_ancestors(minimal=True,
                                   filediffs=self.filediffs)

        with self.assertNumQueries(5):
            for filediff in self.filediffs:
                ancestors[filediff] = filediff.get_ancestors(minimal=True)

        self._check_ancestors(ancestors, minimal=True)

    def _check_ancestors(self, all_ancestors, minimal):
        paths = {
            (1, 'foo', 'PRE-CREATION', 'foo', 'e69de29'): ([], []),
            (1, 'bar', '5716ca5', 'bar', '8e739cc'): ([], []),
            (2, 'foo', 'e69de29', 'foo', '257cc56'): (
                [],
                [
                    (1, 'foo', 'PRE-CREATION', 'foo', 'e69de29'),
                ],
            ),
            (2, 'bar', '8e739cc', 'bar', '0000000'): (
                [],
                [
                    (1, 'bar', '5716ca5', 'bar', '8e739cc'),
                ],
            ),
            (2, 'baz', '7601807', 'baz', '280beb2'): ([], []),
            (3, 'foo', '257cc56', 'qux', '03b37a0'): (
                [],
                [
                    (1, 'foo', 'PRE-CREATION', 'foo', 'e69de29'),
                    (2, 'foo', 'e69de29', 'foo', '257cc56'),
                ],
            ),
            (3, 'bar', 'PRE-CREATION', 'bar', '5716ca5'): (
                [
                    (1, 'bar', '5716ca5', 'bar', '8e739cc'),
                    (2, 'bar', '8e739cc', 'bar', '0000000'),
                ],
                [],
            ),
            (3, 'corge', 'e69de29', 'corge', 'f248ba3'): ([], []),
            (4, 'bar', '5716ca5', 'quux', 'e69de29'): (
                [
                    (1, 'bar', '5716ca5', 'bar', '8e739cc'),
                    (2, 'bar', '8e739cc', 'bar', '0000000'),
                ],
                [
                    (3, 'bar', 'PRE-CREATION', 'bar', '5716ca5'),
                ],
            ),
        }

        by_details = self.get_filediffs_by_details()

        for filediff, ancestors in all_ancestors.items():
            rest_ids, minimal_ids = paths[(
                filediff.commit_id,
                filediff.source_file,
                filediff.source_revision,
                filediff.dest_file,
                filediff.dest_detail,
            )]

            if minimal:
                ids = minimal_ids
            else:
                ids = chain(rest_ids, minimal_ids)

            expected_ancestors = [
                by_details[details] for details in ids
            ]

            self.assertEqual(ancestors, expected_ancestors)



"""Account-related template tags."""

from __future__ import annotations

import logging
from datetime import datetime
from typing import Iterator, TYPE_CHECKING

import pytz
from django import template
from django.contrib.auth.models import User
from django.utils import dateformat, timezone
from django.utils.html import escape, format_html, format_html_join
from django.utils.safestring import SafeString, mark_safe
from django.utils.translation import gettext as _
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.templatetags.djblets_js import json_dumps

from reviewboard.accounts.user_details import user_details_provider_registry
from reviewboard.admin.read_only import is_site_read_only_for
from reviewboard.avatars import avatar_services
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from django.template.context import Context

    from reviewboard.accounts.models import Profile


logger = logging.getLogger(__name__)
register = template.Library()


@register.simple_tag(takes_context=True)
def user_profile_display_name(context, user):
    """Render the user's display name.

    Args:
        context (django.template.context.Context):
            The template rendering context.

        user (django.contrib.auth.models.User):
            The user whose display name is to be rendered.

    Returns:
        unicode:
        The user's display name.
    """
    request = context['request']

    if request is not None:
        request_user = request.user
    else:
        request_user = None

    return escape(user.get_profile().get_display_name(request_user))


@register.simple_tag(takes_context=True)
def js_user_session_info(context):
    """Return JSON-serialized data for a new RB.UserSession instance.

    This is used on all Review Board pages to construct the attributes
    passed to :js:class:`RB.UserSession`. This contains information on the user
    and their authentication state, important API URLs that need to be
    accessed, timezone information, avatar URLs, and various user preferences.

    Args:
        context (django.template.Context):
            The current template context.

    Returns:
        django.utils.safestring.SafeText:
        The JSON-serialized attribute data.
    """
    request = context['request']
    user = request.user
    authenticated = user.is_authenticated

    info = {
        'authenticated': authenticated,
    }

    if authenticated:
        # Authenticated users.
        siteconfig = SiteConfiguration.objects.get_current()
        profile: Profile = request.user.get_profile()
        username = user.username
        avatar_urls = {}
        avatar_html = {}

        info.update({
            'allowSelfShipIt': siteconfig.get('reviews_allow_self_shipit'),
            'fullName': user.get_full_name() or username,
            'readOnly': is_site_read_only_for(user),
            'username': username,
        })

        # Inject some URLs needed to manage some user state.
        info['sessionURL'] = local_site_reverse('session-resource',
                                                request=request)

        for key, url_name in (('archivedReviewRequestsURL',
                               'archived-review-requests-resource'),
                              ('mutedReviewRequestsURL',
                               'muted-review-requests-resource'),
                              ('userFileAttachmentsURL',
                               'user-file-attachments-resource'),
                              ('userPageURL',
                               'user'),
                              ('watchedReviewGroupsURL',
                               'watched-review-groups-resource'),
                              ('watchedReviewRequestsURL',
                               'watched-review-requests-resource')):
            info[key] = local_site_reverse(
                url_name,
                request=request,
                kwargs={
                    'username': username,
                })

        if profile is not None:
            cur_timezone = pytz.timezone(profile.timezone)
            use_rich_text = profile.should_use_rich_text
            info.update({
                'commentsOpenAnIssue': profile.open_an_issue,
                'confirmShipIt': profile.should_confirm_ship_it,
                'enableDesktopNotifications':
                    profile.should_enable_desktop_notifications,
                'enableSpellChecking': profile.should_enable_spell_checking,
                'quickAccessActionIDs': profile.quick_access_actions,
            })
        else:
            cur_timezone = timezone.get_current_timezone()
            use_rich_text = siteconfig.get('default_use_rich_text')

        if siteconfig.get('avatars_enabled'):
            avatar_service = avatar_services.for_user(user)

            if avatar_service is None:
                logger.error('Could not get a suitable avatar service for '
                             'user %s in js_session_info().',
                             user,
                             extra={'request': request})
            else:
                # Fetch a 32x32 avatar URL (and any variants for different
                # screen DPIs). We only fetch 32x32 for historical reasons,
                # but may want to extend this in the future for additional
                # sizes.
                for size in (32,):
                    avatar_urls[size] = avatar_service.get_avatar_urls(
                        request=request,
                        user=user,
                        size=size)
                    avatar_html[size] = avatar_service.render(
                        request=request,
                        user=user,
                        size=size)

        info.update({
            'avatarHTML': avatar_html,
            'avatarURLs': avatar_urls,
            'defaultUseRichText': use_rich_text,
            'timezoneOffset': dateformat.format(datetime.now(tz=cur_timezone),
                                                'O'),
        })
    else:
        # Anonymous users.
        info['loginURL'] = local_site_reverse('login',
                                              request=request)

    return json_dumps(info)


@register.simple_tag(takes_context=True)
def user_badges(
    context: Context,
    user: User,
) -> SafeString:
    """Return badges shown for a user.

    This will query for any badges provided for a user and return HTML
    rendering them. No order is guaranteed.

    Version Added:
        8.0

    Args:
        context (django.template.context.Context):
            The template rendering context.

        user (django.contrib.auth.models.User):
            The user to query for badges.

    Returns:
        django.utils.safestring.SafeString:
        The HTML containing badges for the user.
    """
    if not user:
        return mark_safe('')

    assert isinstance(user, User), f'{user!r} must be a User'

    def _iter_badges() -> Iterator[SafeString]:
        request = context.get('request')

        if request is not None:
            local_site = getattr(request, 'local_site', None)
        else:
            local_site = None

        for provider in user_details_provider_registry:
            try:
                badges = provider.get_user_badges(user=user,
                                                  local_site=local_site,
                                                  request=request)

                for user_badge in badges:
                    yield user_badge.render_to_string()
            except Exception as e:
                logger.exception(
                    'Unexpected error when fetching user badges from provider '
                    '%r: %s',
                    provider, e)

    badges_html = format_html_join('', '{}', (
        (html,)
        for html in _iter_badges()
    ))

    if badges_html:
        return format_html(
            '<span class="rb-c-user-badges"'
            ' aria-label="{label}"'
            ' role="list">{badges_html}</span>',
            label=_('Badges'),
            badges_html=badges_html,
        )
    else:
        return mark_safe('')



"""Console functionality for command line tools."""

import codecs
import getpass
import shutil
import sys
import textwrap

from django.core.exceptions import ValidationError
from django.core.management.color import color_style, no_style
from django.utils import termcolors


_console = None


class _StyledWrapperIndent:
    """Wraps styled text to help provide textwrap with indent lengths.

    :py:mod:`textwrap` calculates the width of the indent strings to figure
    out how many characters can fit on a line, but it's completely unaware of
    ANSI characters. It reduces the available width on the line by the number
    of invisible ANSI control characters.

    This helps wrap and style indent text in a way that allows text wrapping
    to use the actual visible width of the indent rather than the raw
    character width.

    Version Added:
        5.0
    """

    def __init__(self, text, style):
        """Initialize the indent text.

        Args:
            text (str):
                The text to style.

            style (callable):
                The function used to style the text, or ``None`` if the text
                will not be styled.
        """
        self._length = len(text)

        if style:
            self._text = style(text)
        else:
            self._text = text

    def __add__(self, other):
        """Add another string to the contained string.

        This will be used when building the content for a line.

        Args:
            other (str):
                The string to add onto the contained stringr.

        Returns:
            str:
            The combined length.
        """
        return str(self._text) + other

    def __len__(self):
        """Return the visible length of the string.

        This will be used for available text width calculations.

        Returns:
            int:
            The visible length of the string.
        """
        return self._length


class Console(object):
    """Utilities for displaying output to the console.

    This takes care of cleanly outputting various forms of content (text,
    notes, warnings, errors, itemized lists, and more) to the console.

    Callers should construct this by calling :py:func:`init_console`.

    Version Added:
        4.0
    """

    #: Standard text prompts.
    PROMPT_TYPE_TEXT = 'text'

    #: Password prompts.
    PROMPT_TYPE_PASSWORD = 'password'

    #: Yes/No prompts.
    PROMPT_TYPE_YES_NO = 'yes_no'

    def __init__(self, allow_color=False, stdout=sys.stdout,
                 stderr=sys.stderr, default_text_padding=0):
        """Initialize the console.

        Args:
            allow_color (bool, optional):
                Whether to use color in any console output. This applies to
                headers, notes, warnings, errors, and progress steps.

            stdout (io.IOBase):
                The stream to output standard text to.

            stderr (io.IOBase):
                The stream to output errors to.
        """
        writer = codecs.getwriter('utf-8')

        self._allow_color = allow_color

        try:
            # sys.stdout with Python 3.
            self.stdout = writer(stdout.buffer)
        except AttributeError:
            # sys.stdout with Python 2, or some other stream.
            self.stdout = writer(stdout)

        try:
            # sys.stderr with Python 3.
            self.stderr = writer(stderr.buffer)
        except AttributeError:
            # sys.stderr with Python 2, or some other stream.
            self.stderr = writer(stderr)

        self.default_text_padding = default_text_padding

        # Get the terminal width in order to best fit wrapped content.
        term_width = 79

        if hasattr(shutil, 'get_terminal_size'):
            try:
                # get_terminal_size can report 0, 0 if run from pseudo-terminal
                term_width = shutil.get_terminal_size()[0] or term_width
            except OSError:
                pass

        self.term_width = term_width
        self.header_sep = '\u2014' * self.term_width

        self._restyle_console()

    @property
    def allow_color(self):
        """Whether to use color for output.

        Type:
            bool
        """
        return self._allow_color

    @allow_color.setter
    def allow_color(self, allow_color):
        if self._allow_color is not allow_color:
            self._allow_color = allow_color
            self._restyle_console()

    def make_text_wrapper(self, prefix='', prefix_style=None, left_padding=0,
                          right_padding=None):
        """Return a new TextWrapper.

        The resulting :py:class:`textwrap.TextWrapper` will be tailored to the
        terminal width, and will make use of any provided prefix, style, and
        padding.

        Args:
            prefix (unicode, optional):
                A prefix for the first line in the wrapped content.

            prefix_style (callable, optional):
                The style function used to style the prefix.

            left_padding (int, optional):
                Padding to apply to the left of all lines.

            right_padding (int, optional):
                Padding to apply to the right of all lines. This defaults to
                the value of ``left_padding``.

        Returns:
            textwrap.TextWrapper:
            The resulting text wrapper.
        """
        left_indent_len = left_padding + len(prefix)

        if right_padding is None:
            right_padding = left_padding

        return textwrap.TextWrapper(
            initial_indent=_StyledWrapperIndent(
                '%s%s' % (' ' * left_padding, prefix),
                prefix_style),
            subsequent_indent=' ' * left_indent_len,
            break_long_words=False,
            width=self.term_width - right_padding)

    def wrap_text(self, text, indent=None, wrapper=None):
        """Return a paragraph of text wrapped to the terminal width.

        Args:
            text (unicode):
                The text to wrap.

            indent (unicode, optional):
                A custom indentation string.

            wrapper (textwrap.TextWrapper, optional):
                A specific text wrapper to use. Defaults to the standard
                text wrapper for the console.

        Returns:
            unicode:
            The wrapped text.
        """
        if wrapper is None:
            wrapper = self.text_wrapper

        if indent is None:
            result = wrapper.fill(text)
        else:
            old_initial_indent = wrapper.initial_indent
            old_subsequent_indent = wrapper.subsequent_indent
            old_width = wrapper.width

            wrapper.initial_indent = indent
            wrapper.subsequent_indent = indent
            wrapper.width = self.term_width

            result = wrapper.fill(text)

            wrapper.initial_indent = old_initial_indent
            wrapper.subsequent_indent = old_subsequent_indent
            wrapper.width = old_width

        return result

    def print(self, text='', wrap=True, wrapper=None, style=None,
              trailing_newline=True):
        """Display a block of text to the user.

        Args:
            text (unicode):
                The text to display.

            wrap (bool, optional):
                Whether to wrap the text. Any newlines will result in new
                paragraphs.

            wrapper (textwrap.TextWrapper, optional):
                A specific text wrapper to use. Defaults to the standard
                text wrapper for the console.

            style (callable, optional):
                The style function used to style the text.

            trailing_newline (bool, optional):
                Whether to include a trailing newline at the end.
        """
        if style is None:
            style = self._plain_style

        if wrap:
            if wrapper is None:
                wrapper = self.text_wrapper

            indent = None

            for i, paragraph in enumerate(text.strip().splitlines()):
                if i > 0:
                    self.stdout.write('\n\n')
                    indent = wrapper.subsequent_indent

                self.stdout.write(style(self.wrap_text(paragraph,
                                                       wrapper=wrapper,
                                                       indent=indent)))
        else:
            for line in text.splitlines(True):
                self.stdout.write('%s%s' % (' ' * self.default_text_padding,
                                            style(line)))

        if trailing_newline:
            self.stdout.write('\n')

    def note(self, text, leading_newlines=True, trailing_newlines=True):
        """Display a block containing an important note.

        Args:
            text (unicode):
                The text to display.

            leading_newlines (bool, optional):
                Whether to show 2 newlines before the text.

            trailing_newlines (bool, optional):
                Whether to show 1 newline after the text.
        """
        if leading_newlines:
            self.print()

        self.print(text,
                   wrapper=self.note_wrapper)

        if trailing_newlines:
            self.print()

    def warning(self, text, leading_newlines=True, trailing_newlines=True):
        """Display a block containing a warning.

        Args:
            text (unicode):
                The text to display.

            leading_newlines (bool, optional):
                Whether to show 2 newlines before the text.

            trailing_newlines (bool, optional):
                Whether to show 1 newline after the text.
        """
        if leading_newlines:
            self.print()

        self.print(text,
                   wrapper=self.warning_wrapper)

        if trailing_newlines:
            self.print()

    def error(self, text, leading_newlines=True, trailing_newlines=True):
        """Display a block containing a warning.

        Args:
            text (unicode):
                The text to display.

            leading_newlines (bool, optional):
                Whether to show 2 newlines before the text.

            trailing_newlines (bool, optional):
                Whether to show 1 newline after the text.
        """
        if leading_newlines:
            self.print()

        self.print(text,
                   wrapper=self.error_wrapper)

        if trailing_newlines:
            self.print()

    def header(self, title, leading_newlines=True, trailing_newlines=True):
        """Display a header.

        Args:
            title (unicode):
                The header title.

            leading_newlines (bool, optional):
                Whether to show 2 newlines before the header.

            trailing_newlines (bool, optional):
                Whether to show 1 newline after the header.
        """
        if leading_newlines:
            self.print()
            self.print()

        self.print(self.header_sep,
                   style=self.header_sep_style,
                   wrapper=self.header_sep_wrapper)
        self.print(title,
                   style=self.header_sep_style,
                   wrapper=self.header_wrapper)
        self.print(self.header_sep,
                   style=self.header_sep_style,
                   wrapper=self.header_sep_wrapper)

        if trailing_newlines:
            self.print()

    def prompt_input(self, prompt, prompt_type=PROMPT_TYPE_TEXT,
                     default=None, optional=False, strip=True,
                     validate_func=None):
        """Prompt the user for input.

        Args:
            prompt (unicode):
                The text prompting for input.

            prompt_type (unicode, optional):
                The type of input to prompt for. This is one of:

                * :py:attr:`PROMPT_TYPE_TEXT`
                * :py:attr:`PROMPT_TYPE_PASSWORD`
                * :py:attr:`PROMPT_TYPE_YES_NO`

            default (bool or unicode, optional):
                The default value to show and use, if an explicit value isn't
                provided by the user.

                For yes/no prompts, this should be a boolean. For all else,
                a string.

            optional (bool, optional):
                Whether the prompt is optional and can be skipped by omitting
                a value.

            strip (bool, optional):
                Whether to strip the provided input.

            validate_func (callable, optional):
                An optional function for determining if input is valid. This
                takes the input as a parameter and raises a
                :py:class:`django.core.exceptions.ValidationError` if invalid.

                .. code-block:: python

                   def _is_valid(value):
                       if value != 'abc':
                           raise ValidationError('bad!')

        Returns:
            unicode:
            The resulting inputted value.
        """
        if prompt_type == self.PROMPT_TYPE_YES_NO:
            if default is True:
                prompt = '%s [Y/n]' % prompt
            elif default is False:
                prompt = '%s [y/N]' % prompt
                default = False
            else:
                prompt = '%s [y/n]' % prompt
        elif default:
            self.print()
            self.print('The default is "%s"' % default)
            prompt = '%s [%s]' % (prompt, default)
        elif optional:
            prompt = '%s (optional)' % prompt

        self.print()

        prompt = self.prompt_style('%s: ' % prompt)
        value = None

        while not value:
            self.print(prompt, trailing_newline=False)
            self.stdout.flush()

            if prompt_type == self.PROMPT_TYPE_PASSWORD:
                value = getpass.getpass(str(''),
                                        stream=self.stdout)
            else:
                value = input()

            if strip:
                value = value.strip()

            if not value:
                if default:
                    value = default
                elif optional:
                    break

            if validate_func is not None:
                try:
                    validate_func(value)
                except ValidationError as e:
                    for error in e.messages:
                        self.error(error)

                    value = None
                    continue

            if prompt_type == self.PROMPT_TYPE_YES_NO:
                if isinstance(value, bool):
                    # This came from the 'default' value.
                    norm_value = value
                else:
                    assert isinstance(value, str)
                    norm_value = value.lower()

                if norm_value not in (True, False, 'y', 'n', 'yes', 'no'):
                    self.error('Must specify one of Y/y/yes or N/n/no.')
                    value = None
                    continue
                else:
                    value = norm_value in (True, 'y', 'yes')
                    break
            elif not value:
                self.error('An answer is required.')

        return value

    def prompt_choice(self, prompt, choices):
        """Prompt the user for a choice from a list.

        Args:
            prompt (unicode):
                The text prompting for a choice.

            choices (list of dict):
                The list of choices to present. Each entry is a dictionary
                with the following keys:

                ``text`` (:py:class:`unicode`):
                    The text for the choice.

                ``description`` (:py:class:`unicode`, optional):
                    A description of the choice.

                ``enabled`` (:py:class:`bool`, optional):
                    Whether the option is enabled/visible.

        Returns:
            object:
            The resulting choice.
        """
        self.print()
        self.print('You can type either the name or the number from the '
                   'list below.')
        self.print()

        prompt_style = self.prompt_style
        valid_choices = []
        i = 0

        for choice in choices:
            if choice.get('enabled', True):
                text = choice['text']

                self.print(
                    '%s %s %s\n'
                    % (prompt_style('(%d)' % (i + 1)),
                       text,
                       choice.get('description', '')))
                valid_choices.append(text)
                i += 1

        self.print()

        prompt = self.prompt_style('%s: ' % prompt)
        choice = None

        while not choice:
            self.print(prompt, trailing_newline=False)
            choice = input()

            if choice not in valid_choices:
                try:
                    i = int(choice) - 1

                    if 0 <= i < len(valid_choices):
                        choice = valid_choices[i]
                        break
                except ValueError:
                    pass

                self.error("'%s' is not a valid option." % choice)
                choice = None

        return choice

    def itemized_list(self, items, title=''):
        """Display a list of items.

        Args:
            items (list of unicode):
                The list of items to show.

            title (unicode, optional):
                An optional title to show above the list.
        """
        self.print()

        if title:
            self.print('%s:' % title)
            self.print()

        wrapper = self.item_wrapper

        for item in items:
            self.print(item, wrapper=wrapper)

    def progress_step(self, text, func, step_num=None, total_steps=None):
        """Display one step of a multi-step operation.

        This will indicate when it's starting and when it's complete.

        If both ``step_num`` and ``total_steps`` are provided, the step
        text will include a prefix showing what step it's on and how many
        there are total.

        Args:
            text (unicode):
                The step text to display.

            func (callable):
                The function to call to execute the step.

            step_num (int, optional):
                The 1-based step number.

            total_steps (int, optional):
                The total number of steps.
        """
        assert callable(func)

        if step_num is not None and total_steps is not None:
            text = '[%s/%s] %s' % (step_num, total_steps, text)

        self.print('%s ... ' % text,
                   trailing_newline=False,
                   wrap=False)

        try:
            func()
            self.stdout.write(self.style.SUCCESS('OK'))
        except Exception as e:
            self.stdout.write('%s %s' % (self.style.ERROR('ERROR:'), e))

        self.stdout.write('\n')

    def _plain_style(self, text):
        """Return text as-is, without any styling.

        Args:
            text (unicode):
                The text to "style".

        Returns:
            unicode:
            The provided text.
        """
        return text

    def _restyle_console(self):
        """Restyle console output.

        This will create/re-create the output styles, based on the terminal
        size and whether color is allowed.
        """
        # Recompute the styles, based on whether color is allowed.
        if self.allow_color:
            self.style = color_style()

            self.header_style = termcolors.make_style(fg='yellow',
                                                      bg='black',
                                                      opts=('bold',))
            self.header_sep_style = termcolors.make_style(fg='yellow',
                                                          bg='black')
            self.prompt_style = termcolors.make_style(opts=('bold',))
        else:
            self.style = no_style()

            plain_style = self._plain_style
            self.header_style = plain_style
            self.header_sep_style = plain_style
            self.prompt_style = plain_style

        # Rebuild the text wrappers.
        text_padding = self.default_text_padding

        self.header_wrapper = self.make_text_wrapper(
            left_padding=1,
            right_padding=1)
        self.header_sep_wrapper = self.make_text_wrapper()
        self.text_wrapper = self.make_text_wrapper(
            left_padding=text_padding,
            right_padding=text_padding)
        self.note_wrapper = self.make_text_wrapper(
            prefix='Note: ',
            prefix_style=self.style.WARNING,
            left_padding=text_padding,
            right_padding=text_padding)
        self.warning_wrapper = self.make_text_wrapper(
            prefix='Warning: ',
            prefix_style=self.style.WARNING,
            left_padding=text_padding,
            right_padding=text_padding)
        self.error_wrapper = self.make_text_wrapper(
            prefix='[!] ',
            prefix_style=self.style.ERROR,
            left_padding=text_padding,
            right_padding=text_padding)
        self.item_wrapper = self.make_text_wrapper(
            prefix='* ',
            left_padding=text_padding,
            right_padding=text_padding)


def init_console(*args, **kwargs):
    """Initialize the console.

    This can only be called once.

    Args:
        **kwargs (dict):
            Keyword arguments to pass to :py:class:`Console`.

    Returns:
        Console:
        The resulting console instance.
    """
    global _console
    assert _console is None, 'init_console() was already called.'

    _console = Console(*args, **kwargs)

    return _console


def uninit_console():
    """Uninitialize the console."""
    global _console
    assert _console is not None, 'init_console() was never called.'

    _console = None


def get_console():
    """Return the console instance.

    Returns:
        Console:
        The initialized console, or ``None`` if not yet initialized.
    """
    return _console



"""Unit tests for the Forgejo hosting service.

Version Added:
    8.0
"""

from __future__ import annotations

import base64
import hashlib
import hmac
import json
from typing import TYPE_CHECKING

import kgb
from djblets.testing.decorators import add_fixtures

from reviewboard.hostingsvcs.errors import (
    AuthorizationError,
    HostingServiceError,
    TwoFactorAuthCodeRequiredError,
)
from reviewboard.hostingsvcs.forgejo import Forgejo
from reviewboard.hostingsvcs.forgejo.client import ForgejoClient
from reviewboard.hostingsvcs.forgejo.errors import APITokenNameInUseError
from reviewboard.hostingsvcs.hook_utils import logger
from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.reviews.models import ReviewRequest
from reviewboard.scmtools.crypto_utils import (
    decrypt_password,
    encrypt_password,
)
from reviewboard.scmtools.errors import FileNotFoundError
from reviewboard.site.models import LocalSite
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from django.http import HttpResponse


class ForgejoTestCase(HostingServiceTestCase[Forgejo]):
    """Base class for Forgejo test suites.

    Version Added:
        8.0
    """

    service_name = 'forgejo'

    default_account_data = {
        'api_token': encrypt_password('test_token_123'),
    }

    default_repository_extra_data = {
        'repository_owner': 'testuser',
        'repository_name': 'testrepo',
    }

    default_hosting_url = 'https://forge.example.com'
    default_use_hosting_url = True


class ForgejoTests(ForgejoTestCase):
    """Unit tests for the Forgejo hosting service.

    Version Added:
        8.0
    """

    def test_service_support(self) -> None:
        """Testing Forgejo service support capabilities"""
        service_class = self.service_class
        assert service_class is not None
        self.assertTrue(service_class.supports_repositories)
        self.assertTrue(service_class.supports_post_commit)
        self.assertTrue(service_class.supports_two_factor_auth)
        self.assertTrue(service_class.self_hosted)
        self.assertEqual(service_class.supported_scmtools, ['Git'])

    def test_get_repository_fields(self) -> None:
        """Testing Forgejo.get_repository_fields"""
        fields = self.get_repository_fields(
            'Git',
            fields={
                'repository_owner': 'myowner',
                'repository_name': 'myrepo',
            }
        )

        self.assertEqual(fields, {
            'path': 'https://forge.example.com/myowner/myrepo.git',
        })

    def test_is_authorized_with_token(self) -> None:
        """Testing Forgejo.is_authorized with API token"""
        account = self.create_hosting_account()
        service = account.service

        self.assertTrue(service.is_authorized())

    def test_is_authorized_without_token(self) -> None:
        """Testing Forgejo.is_authorized without API token"""
        account = self.create_hosting_account(data={})
        service = account.service

        self.assertFalse(service.is_authorized())

    def test_authorize_success(self) -> None:
        """Testing Forgejo.authorize with valid credentials"""
        access_token_data = {
            'id': 1,
            'name': 'Review Board',
            'scopes': ['read:issue', 'read:organization', 'read:repository',
                       'read:user'],
            'sha1': 'new_test_token_456',
            'token_last_eight': '456',
        }

        account = self.create_hosting_account(data={})
        service = account.service

        self.assertFalse(service.is_authorized())

        handler = self.make_handler_for_paths({
            '/api/v1/users/myuser/tokens': {
                'payload': self.dump_json(access_token_data),
                'headers': {'Content-Type': 'application/json'},
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1):
            service.authorize(
                username='myuser',
                password='testpass',
                hosting_url='https://forge.example.com',
                credentials={}
            )

        self.assertTrue(service.is_authorized())
        self.assertEqual(decrypt_password(account.data['api_token']),
                         'new_test_token_456')

    def test_authorize_with_two_factor_auth(self) -> None:
        """Testing Forgejo.authorize with two-factor authentication code"""
        access_token_data = {
            'id': 1,
            'name': 'Review Board',
            'scopes': ['read:issue', 'read:organization', 'read:repository',
                       'read:user'],
            'sha1': 'new_test_token_456',
            'token_last_eight': '456',
        }

        account = self.create_hosting_account(data={})
        service = account.service

        handler = self.make_handler_for_paths({
            '/api/v1/users/myuser/tokens': {
                'payload': self.dump_json(access_token_data),
                'headers': {'Content-Type': 'application/json'},
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1):
            service.authorize(
                username='myuser',
                password='testpass',
                hosting_url='https://forge.example.com',
                credentials={},
                two_factor_auth_code='123456'
            )

        self.assertTrue(service.is_authorized())

    def test_authorize_invalid_credentials(self) -> None:
        """Testing Forgejo.authorize with invalid credentials"""
        error_data = {
            'message': 'invalid provided password',
            'url': 'https://forge.example.com/api/v1/users/myuser/tokens',
        }

        account = self.create_hosting_account(data={})
        service = account.service

        handler = self.make_handler_for_paths({
            '/api/v1/users/myuser/tokens': {
                'payload': self.dump_json(error_data),
                'status_code': 401,
                'headers': {'Content-Type': 'application/json'},
            },
        })

        message = 'username or password is incorrect'

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1), \
             self.assertRaisesMessage(AuthorizationError,
                                      message):
            service.authorize(
                username='myuser',
                password='wrongpass',
                hosting_url='https://forge.example.com',
                credentials={}
            )

    def test_authorize_two_factor_required(self) -> None:
        """Testing Forgejo.authorize when two-factor auth is required"""
        error_data = {
            'message': 'invalid provided OTP',
            'url': 'https://forge.example.com/api/v1/users/myuser/tokens',
        }

        account = self.create_hosting_account(data={})
        service = account.service

        handler = self.make_handler_for_paths({
            '/api/v1/users/myuser/tokens': {
                'payload': self.dump_json(error_data),
                'status_code': 400,
                'headers': {'Content-Type': 'application/json'},
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1), \
             self.assertRaises(TwoFactorAuthCodeRequiredError):
            service.authorize(
                username='myuser',
                password='testpass',
                hosting_url='https://forge.example.com',
                credentials={}
            )

    def test_authorize_with_token_name_retry_success(self) -> None:
        """Testing Forgejo.authorize with successful retry after duplicate
        token name
        """
        account = self.create_hosting_account(data={})
        service = account.service
        client = service.client
        assert isinstance(client, ForgejoClient)

        self.spy_on(
            client.create_api_token,
            op=kgb.SpyOpMatchInOrder([
                {
                    'op': kgb.SpyOpRaise(APITokenNameInUseError()),
                },
                {
                    'op': kgb.SpyOpReturn('retry_success_token_456'),
                }
            ]))

        service.authorize(
            username='myuser',
            password='testpass',
            hosting_url='https://forge.example.com',
            credentials={}
        )

        self.assertTrue(service.is_authorized())
        self.assertEqual(decrypt_password(account.data['api_token']),
                         'retry_success_token_456')

    def test_authorize_with_max_retries_exceeded(self) -> None:
        """Testing Forgejo.authorize fails after maximum retries due to
        duplicate token names
        """
        account = self.create_hosting_account(data={})
        service = account.service
        client = service.client
        assert isinstance(client, ForgejoClient)

        self.spy_on(client.create_api_token,
                    op=kgb.SpyOpRaise(APITokenNameInUseError()))

        message = (
            'Unable to create a Forgejo API token with a unique name after 5 '
            'attempts for user myuser on server https://forge.example.com'
        )

        with self.assertRaisesMessage(HostingServiceError, message):
            service.authorize(
                username='myuser',
                password='testpass',
                hosting_url='https://forge.example.com',
                credentials={}
            )

        self.assertFalse(service.is_authorized())

        spy = self.get_spy(client.create_api_token)
        self.assertEqual(len(spy.calls), 5)

    def test_create_api_token_name(self) -> None:
        """Testing Forgejo._create_api_token_name generates unique names"""
        account = self.create_hosting_account(data={})
        service = account.service
        assert isinstance(service, Forgejo)

        # Test that token names are generated with correct format
        token_name1 = service._create_api_token_name('example.com')
        token_name2 = service._create_api_token_name('example.com')

        # Both should start with 'reviewboard-example.com-'
        self.assertTrue(token_name1.startswith('reviewboard-example.com-'))
        self.assertTrue(token_name2.startswith('reviewboard-example.com-'))

        # They should be different (due to UUID)
        self.assertNotEqual(token_name1, token_name2)

        # Both should have the expected structure
        parts1 = token_name1.split('-')
        parts2 = token_name2.split('-')

        self.assertEqual(len(parts1), 3)
        self.assertEqual(len(parts2), 3)

        self.assertEqual(parts1[0], 'reviewboard')
        self.assertEqual(parts1[1], 'example.com')
        self.assertEqual(len(parts1[2]), 6)  # UUID hex should be 6 chars

        self.assertEqual(parts2[0], 'reviewboard')
        self.assertEqual(parts2[1], 'example.com')
        self.assertEqual(len(parts2[2]), 6)

    def test_create_api_token_name_truncates_long_server_name(self) -> None:
        """Testing Forgejo._create_api_token_name truncates server names to
        stay within 255 char limit
        """
        account = self.create_hosting_account(data={})
        service = account.service
        assert isinstance(service, Forgejo)

        long_server_name = ('a' * 250) + '.com'

        token_name = service._create_api_token_name(long_server_name)
        self.assertEqual(len(token_name), 255)

    def test_get_file(self) -> None:
        """Testing Forgejo.get_file"""
        file_content = b'def main():\n    print("Hello, World!")\n'
        encoded_content = base64.b64encode(file_content).decode('utf-8')

        blob_data = {
            'content': encoded_content,
            'encoding': 'base64',
            'sha': 'abc123def456',
            'size': len(file_content),
            'url': 'https://forge.example.com/api/v1/repos/testuser/testrepo/'
                   'git/blobs/abc123def456',
        }

        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        service = account.service

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo/git/blobs/abc123def456': {
                'payload': self.dump_json(blob_data),
                'headers': {'Content-Type': 'application/json'},
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1):
            content = service.get_file(
                repository=repository,
                path='main.py',
                revision='abc123def456'
            )

        self.assertEqual(content, file_content)

    def test_get_file_not_found(self) -> None:
        """Testing Forgejo.get_file with file not found"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        service = account.service

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo/git/blobs/abc123def456': {
                'status_code': 400,
                'payload': json.dumps({
                    'message': 'object does not exist [id: abc123def456, '
                               'rel_path: ]',
                    'url': '/api/v1/repos/testuser/testrepo/git/blobs/'
                           'abc123def456',
                }).encode(),
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1), \
             self.assertRaises(FileNotFoundError):
            service.get_file(
                repository=repository,
                path='nonexistent.py',
                revision='abc123def456'
            )

    def test_get_file_repo_invalid(self) -> None:
        """Testing Forgejo.get_file with an invalid repository name"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        service = account.service

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo/git/blobs/abc123def456': {
                'status_code': 404,
                'payload': json.dumps({
                    'errors': [],
                    'message': "The target couldn't be found.",
                    'url': '/api/v1/repos/testuser/testrepo/git/blobs/'
                           'abc123def456',
                }).encode(),
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1), \
             self.assertRaises(HostingServiceError):
            service.get_file(
                repository=repository,
                path='nonexistent.py',
                revision='abc123def456'
            )

    def test_get_branches(self) -> None:
        """Testing Forgejo.get_branches"""
        repo_data = {
            'clone_url': 'https://forge.example.com/testuser/testrepo.git',
            'default_branch': 'main',
            'description': 'My repo',
            'id': 1,
            'name': 'testrepo',
            'private': False,
        }

        refs_data = [
            {
                'object': {
                    'sha': 'abc123',
                    'type': 'commit',
                    'url': 'https://forge.example.com/api/v1/repos/testuser/'
                           'testrepo/git/commits/abc123',
                },
                'ref': 'refs/heads/main',
                'url': 'https://forge.example.com/api/v1/repos/testuser/'
                       'testrepo/git/refs/heads/main',
            },
            {
                'object': {
                    'sha': 'def456',
                    'type': 'commit',
                    'url': 'https://forge.example.com/api/v1/repos/testuser/'
                           'testrepo/git/commits/def456',
                },
                'ref': 'refs/heads/feature',
                'url': 'https://forge.example.com/api/v1/repos/testuser/'
                       'testrepo/git/refs/heads/feature',
            }
        ]

        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        service = account.service

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo': {
                'payload': self.dump_json(repo_data),
                'headers': {'Content-Type': 'application/json'},
            },
            '/api/v1/repos/testuser/testrepo/git/refs': {
                'payload': self.dump_json(refs_data),
                'headers': {'Content-Type': 'application/json'},
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=2):
            branches = service.get_branches(repository)

        self.assertEqual(len(branches), 2)

        main_branch = next(b for b in branches if b.id == 'main')
        self.assertEqual(main_branch.commit, 'abc123')
        self.assertTrue(main_branch.default)

        feature_branch = next(b for b in branches if b.id == 'feature')
        self.assertEqual(feature_branch.commit, 'def456')
        self.assertFalse(feature_branch.default)

    def test_get_commits(self) -> None:
        """Testing Forgejo.get_commits"""
        commits_data = [
            {
                'commit': {
                    'author': {
                        'date': '2023-01-01T12:00:00Z',
                        'email': 'user@example.com',
                        'name': 'Test User',
                    },
                    'committer': {
                        'date': '2023-01-01T12:00:00Z',
                        'email': 'user@example.com',
                        'name': 'Test User',
                    },
                    'message': 'Initial commit',
                    'tree': {
                        'created': '2023-01-01T12:00:00Z',
                        'sha': 'tree123',
                        'url': 'https://forge.example.com/api/v1/repos/'
                               'testuser/testrepo/git/trees/tree123',
                    }
                },
                'created': '2023-01-01T12:00:00Z',
                'parents': [],
                'sha': 'commit123',
            }
        ]

        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        service = account.service

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo/commits': {
                'payload': self.dump_json(commits_data),
                'headers': {'Content-Type': 'application/json'},
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1):
            commits = service.get_commits(repository, branch='main')

        self.assertEqual(len(commits), 1)
        commit = commits[0]
        self.assertEqual(commit.author_name, 'Test User')
        self.assertEqual(commit.id, 'commit123')
        self.assertEqual(commit.message, 'Initial commit')
        self.assertEqual(commit.parent, '')

    def test_get_commits_with_no_start(self) -> None:
        """Testing Forgejo.get_commits with no start point"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        service = account.service

        with self.assertRaises(ValueError):
            service.get_commits(repository)

    def test_get_change(self) -> None:
        """Testing Forgejo.get_change"""
        commit_data = {
            'commit': {
                'author': {
                    'date': '2023-01-01T12:00:00Z',
                    'email': 'user@example.com',
                    'name': 'Test User',
                },
                'committer': {
                    'date': '2023-01-01T12:00:00Z',
                    'email': 'user@example.com',
                    'name': 'Test User',
                },
                'message': 'Add new feature',
                'tree': {
                    'created': '2023-01-01T12:00:00Z',
                    'sha': 'tree123',
                    'url': 'https://forge.example.com/api/v1/repos/testuser/'
                           'testrepo/git/trees/tree123',
                }
            },
            'created': '2023-01-01T12:00:00Z',
            'parents': [
                {
                    'created': '2023-01-01T11:00:00Z',
                    'sha': 'parent123',
                    'url': '/api/v1/repos/testuser/testrepo/git/commits/'
                           'parent123',
                },
            ],
            'sha': 'commit456',
        }

        tree_data = {
            'page': 1,
            'sha': 'tree123',
            'total_count': 1,
            'tree': [
                {
                    'path': 'file.py',
                    'sha': 'file123',
                    'size': 100,
                    'type': 'blob',
                }
            ],
            'truncated': False
        }

        diff_content = (
            b'diff --git a/file.py b/file.py\n'
            b'index abc123..def456 100644\n'
            b'--- a/file.py\n'
            b'+++ b/file.py\n'
            b'@@ -1 +1,2 @@\n'
            b' print("Hello")\n'
            b'+print("World")\n'
        )

        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        service = account.service

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo/git/commits/commit456': {
                'payload': self.dump_json(commit_data),
                'headers': {'Content-Type': 'application/json'},
            },
            '/api/v1/repos/testuser/testrepo/git/commits/commit456.diff': {
                'payload': diff_content,
                'headers': {'Content-Type': 'text/plain'},
            },
            '/api/v1/repos/testuser/testrepo/git/trees/commit456': {
                'payload': self.dump_json(tree_data),
                'headers': {'Content-Type': 'application/json'},
            },
            '/api/v1/repos/testuser/testrepo/git/trees/parent123': {
                'payload': self.dump_json(tree_data),
                'headers': {'Content-Type': 'application/json'},
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=4):
            commit = service.get_change(repository, 'commit456')

        self.assertEqual(commit.author_name, 'Test User')
        self.assertEqual(commit.id, 'commit456')
        self.assertEqual(commit.message, 'Add new feature')
        self.assertEqual(commit.parent, 'parent123')
        self.assertIsNotNone(commit.diff)

    def test_get_bug_info_uncached(self) -> None:
        """Testing Forgejo.get_bug_info_uncached"""
        issue_data = {
            'body': 'This is a test issue description with multiple lines.\n'
                    'It should be properly returned in the bug info.',
            'id': 123,
            'state': 'open',
            'title': 'Test Issue Title',
        }

        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        service = account.service
        assert isinstance(service, Forgejo)

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo/issues/123': {
                'payload': self.dump_json(issue_data),
                'headers': {'Content-Type': 'application/json'},
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1):
            bug_info = service.get_bug_info_uncached(repository, '123')

        self.assertEqual(bug_info, {
            'description':
                'This is a test issue description with multiple lines.\n'
                'It should be properly returned in the bug info.',
            'description_text_format': 'markdown',
            'status': 'open',
            'summary': 'Test Issue Title',
        })

    def test_get_bug_info_uncached_issue_not_found(self) -> None:
        """Testing Forgejo.get_bug_info_uncached with issue not found"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        service = account.service
        assert isinstance(service, Forgejo)

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo/issues/999': {
                'status_code': 404,
                'payload': json.dumps({
                    'message': 'Not Found',
                    'url': '/api/v1/repos/testuser/testrepo/issues/999',
                }).encode(),
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1):
            bug_info = service.get_bug_info_uncached(repository, '999')

        self.assertEqual(bug_info, {
            'description': '',
            'status': '',
            'summary': '',
        })

    def test_get_bug_info_uncached_api_error(self) -> None:
        """Testing Forgejo.get_bug_info_uncached with API error"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        service = account.service
        assert isinstance(service, Forgejo)

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo/issues/456': {
                'status_code': 500,
                'payload': json.dumps({
                    'message': 'Internal server error',
                    'url': '/api/v1/repos/testuser/testrepo/issues/456',
                }).encode(),
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1):
            bug_info = service.get_bug_info_uncached(repository, '456')

        self.assertEqual(bug_info, {
            'description': '',
            'status': '',
            'summary': '',
        })


class ForgejoClientTests(ForgejoTestCase):
    """Unit tests for the ForgejoClient

    Version Added:
        8.0
    """

    def test_get_http_credentials_with_token(self) -> None:
        """Testing ForgejoClient.get_http_credentials with API token"""
        account = self.create_hosting_account()
        client = account.service.client

        credentials = client.get_http_credentials(account)

        expected_token = decrypt_password(account.data['api_token'])
        self.assertEqual(expected_token, 'test_token_123')
        self.assertEqual(credentials, {
            'headers': {
                'Authorization': f'token {expected_token}',
            },
        })

    def test_get_http_credentials_without_token(self) -> None:
        """Testing ForgejoClient.get_http_credentials without API token"""
        account = self.create_hosting_account(data={})
        client = account.service.client

        credentials = client.get_http_credentials(account)

        self.assertEqual(credentials, {})

    def test_create_api_token(self) -> None:
        """Testing ForgejoClient.create_api_token"""
        access_token_data = {
            'id': 1,
            'name': 'Review Board',
            'scopes': ['read:issue', 'read:organization', 'read:repository',
                       'read:user'],
            'sha1': 'new_api_token_789',
            'token_last_eight': '789',
        }

        account = self.create_hosting_account()
        client = account.service.client
        assert isinstance(client, ForgejoClient)

        handler = self.make_handler_for_paths({
            '/api/v1/users/myuser/tokens': {
                'payload': self.dump_json(access_token_data),
                'headers': {'Content-Type': 'application/json'},
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1):
            token = client.create_api_token(
                token_name='Review Board',
                hosting_url='https://forge.example.com',
                username='myuser',
                password='testpass',
            )

        self.assertEqual(token, 'new_api_token_789')

    def test_get_blob(self) -> None:
        """Testing ForgejoClient.get_blob"""
        file_content = b'print("Hello from Forgejo!")'
        encoded_content = base64.b64encode(file_content).decode('utf-8')

        blob_data = {
            'content': encoded_content,
            'encoding': 'base64',
            'sha': 'blob789',
            'size': len(file_content),
            'url': 'https://forge.example.com/api/v1/repos/testuser/testrepo/'
                   'git/blobs/blob789',
        }

        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        client = account.service.client
        assert isinstance(client, ForgejoClient)

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo/git/blobs/blob789': {
                'payload': self.dump_json(blob_data),
                'headers': {'Content-Type': 'application/json'},
            },
        })

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1):
            content = client.get_blob(
                hosting_url='https://forge.example.com',
                repository=repository,
                path='hello.py',
                sha='blob789',
            )

        self.assertEqual(content, file_content)

    def test_get_blob_unsupported_encoding(self) -> None:
        """Testing ForgejoClient.get_blob with unsupported encoding"""
        blob_data = {
            'content': 'raw content',
            'encoding': 'raw',
            'sha': 'blob789',
            'size': 11,
            'url': 'https://forge.example.com/api/v1/repos/testuser/testrepo/'
                   'git/blobs/blob789',
        }

        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        client = account.service.client
        assert isinstance(client, ForgejoClient)

        handler = self.make_handler_for_paths({
            '/api/v1/repos/testuser/testrepo/git/blobs/blob789': {
                'payload': self.dump_json(blob_data),
                'headers': {'Content-Type': 'application/json'},
            },
        })

        message = 'Forgejo returned data with an unknown encoding.'

        with self.setup_http_test(http_request_func=handler,
                                  expected_http_calls=1), \
             self.assertRaisesMessage(HostingServiceError,
                                      message):
            client.get_blob(
                hosting_url='https://forge.example.com',
                repository=repository,
                path='hello.py',
                sha='blob789',
            )

    def test_api_root_generation(self) -> None:
        """Testing ForgejoClient._get_api_root"""
        account = self.create_hosting_account()
        client = account.service.client
        assert isinstance(client, ForgejoClient)

        # Test with trailing slash
        self.assertEqual(
            client._get_api_root('https://forge.example.com/'),
            'https://forge.example.com/api/v1'
        )

        # Test without trailing slash
        self.assertEqual(
            client._get_api_root('https://forge.example.com'),
            'https://forge.example.com/api/v1'
        )

    def test_api_repo_root_generation(self) -> None:
        """Testing ForgejoClient._get_api_repo_root"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        client = account.service.client
        assert isinstance(client, ForgejoClient)

        repo_root = client._get_api_repo_root(
            'https://forge.example.com',
            repository
        )

        self.assertEqual(
            repo_root,
            'https://forge.example.com/api/v1/repos/testuser/testrepo'
        )


class ForgejoCloseSubmittedHookTests(ForgejoTestCase):
    """Unit tests for the Forgejo close-submitted webhook.

    Version Added:
        8.0
    """

    fixtures = ['test_users', 'test_scmtools']

    def test_hook(self) -> None:
        """Testing Forgejo close_submitted hook with event=push"""
        self._test_post_commit_hook()

    @add_fixtures(['test_site'])
    def test_hook_with_local_site(self) -> None:
        """Testing Forgejo close_submitted hook with event=push and using a
        Local Site
        """
        self._test_post_commit_hook(
            LocalSite.objects.get(name=self.local_site_name))

    @add_fixtures(['test_site'])
    def test_hook_with_unpublished_review_request(self) -> None:
        """Testing Forgejo close_submitted hook with event=push and an
        un-published review request
        """
        self._test_post_commit_hook(publish=False)

    def test_hook_with_invalid_repo(self) -> None:
        """Testing Forgejo close_submitted hook with event=push and invalid
        repository
        """
        repository = self.create_repository()

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'forgejo-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'forgejo',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_hook_with_invalid_site(self) -> None:
        """Testing Forgejo close_submitted hook with event=push and invalid
        Local Site
        """
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'forgejo-hooks-close-submitted',
            local_site_name='badsite',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'forgejo',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_hook_with_invalid_service_id(self) -> None:
        """Testing Forgejo close_submitted hook with event=push and invalid
        hosting service ID
        """
        # We'll test against GitHub for this test.
        account = self.create_hosting_account()
        account.service_name = 'github'
        account.save()

        repository = self.create_repository(hosting_account=account)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'forgejo-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'forgejo',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_hook_with_invalid_event(self) -> None:
        """Testing Forgejo close_submitted hook with invalid event"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'forgejo-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'forgejo',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid(),
            event='issues')
        self.assertEqual(response.status_code, 400)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_hook_with_missing_signature(self) -> None:
        """Testing Forgejo close_submitted hook with missing signature"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'forgejo-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'forgejo',
            })

        payload = self.dump_json({
            'ref': 'refs/heads/main',
            'commits': [
                {
                    'id': '1c44b461cebe5874a857c51a4a13a849a4d1e52d',
                    'message': (
                        f'This is my fancy commit\n'
                        f'\n'
                        f'Reviewed at '
                        f'http://example.com'
                        f'{review_request.get_absolute_url()}'
                    ),
                },
            ]
        })

        response = self.client.post(
            url,
            payload,
            content_type='application/json',
            HTTP_X_FORGEJO_EVENT='push')
        self.assertEqual(response.status_code, 400)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_hook_with_invalid_signature(self) -> None:
        """Testing Forgejo close_submitted hook with invalid signature"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'forgejo-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'forgejo',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret='bad-secret')
        self.assertEqual(response.status_code, 400)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_hook_with_invalid_review_requests(self) -> None:
        """Testing Forgejo close_submitted hook with event=push and invalid
        review requests
        """
        self.spy_on(logger.error)

        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'forgejo-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'forgejo',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url='/r/9999/',
            secret=repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 200)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

        self.assertSpyCalledWith(
            logger.error,
            'close_all_review_requests: Review request #%s does not exist.',
            9999)

    def test_hook_with_invalid_json(self) -> None:
        """Testing Forgejo close_submitted hook with invalid JSON payload"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'forgejo-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'forgejo',
            })

        invalid_payload = b'invalid json'
        secret = repository.get_or_create_hooks_uuid()
        m = hmac.new(secret.encode('utf-8'), invalid_payload, hashlib.sha256)

        response = self.client.post(
            url,
            invalid_payload,
            content_type='application/json',
            HTTP_X_FORGEJO_EVENT='push',
            HTTP_X_FORGEJO_SIGNATURE=m.hexdigest())
        self.assertEqual(response.status_code, 400)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def _test_post_commit_hook(
        self,
        local_site: (LocalSite | None) = None,
        publish: bool = True,
    ) -> None:
        """Test posting to a commit hook.

        This will simulate pushing a commit and posting the resulting webhook
        payload from Forgejo to the handler for the hook.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site owning the review request.

            publish (bool, optional):
                Whether to test with a published review request.
        """
        account = self.create_hosting_account(local_site=local_site)
        repository = self.create_repository(hosting_account=account,
                                            local_site=local_site)

        review_request = self.create_review_request(repository=repository,
                                                    local_site=local_site,
                                                    publish=publish)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'forgejo-hooks-close-submitted',
            local_site=local_site,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'forgejo',
            })

        response = self._post_commit_hook_payload(
            post_url=url,
            review_request_url=review_request.get_absolute_url(),
            secret=repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 200)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.SUBMITTED)
        self.assertEqual(review_request.changedescs.count(), 1)

        changedesc = review_request.changedescs.get()
        self.assertEqual(changedesc.text, 'Pushed to main (1c44b46)')

    def _post_commit_hook_payload(
        self,
        post_url: str,
        review_request_url: str,
        secret: str,
        event: str = 'push'
    ) -> HttpResponse:
        """Post a payload for a hook for testing.

        Args:
            post_url (str):
                The URL to post to.

            review_request_url (str):
                The URL of the review request being represented in the
                payload.

            secret (str):
                The HMAC secret for the message.

            event (str, optional):
                The webhook event.

        Returns:
            django.http.HttpResponse:
            The response from the post request.
        """
        payload = json.dumps({
            # NOTE: This payload only contains the content we make
            #       use of in the hook.
            'ref': 'refs/heads/main',
            'commits': [
                {
                    'id': '1c44b461cebe5874a857c51a4a13a849a4d1e52d',
                    'message':
                        f'This is my fancy commit\n'
                        f'\n'
                        f'Reviewed at http://example.com{review_request_url}',
                },
            ]
        }).encode()

        m = hmac.new(secret.encode('utf-8'), payload, hashlib.sha256)

        return self.client.post(  # type:ignore
            post_url,
            payload,
            content_type='application/json',
            HTTP_X_FORGEJO_EVENT=event,
            HTTP_X_FORGEJO_SIGNATURE=m.hexdigest())
